# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/MOD11
"""

# import general python modules
import os
import numpy as np
import pandas as pd
from osgeo import gdal
import urllib
from bs4 import BeautifulSoup
import re
import glob
import requests
import sys
if sys.version_info[0] == 3:
    import urllib.parse
if sys.version_info[0] == 2:
    import urlparse
    import urllib2

def DownloadData(Dir, Startdate, Enddate, latlim, lonlim, username, password, Waitbar, hdf_library, remove_hdf):
    """
    This function downloads MOD11 daily LST data

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax] (values must be between -90 and 90)
    lonlim -- [xmin, xmax] (values must be between -180 and 180)
    Waitbar -- 1 (Default) will print a waitbar
    hdf_library -- string, if all the hdf files are already stored on computer
                    define directory to the data here
    remove_hdf -- 1 (Default), if 1 remove all the downloaded hdf files in the end    
    """
	
    # Check start and end date and otherwise set the date to max
    if not Startdate:
        Startdate = pd.Timestamp('2000-02-18')
    if not Enddate:
        Enddate = pd.Timestamp('Now')

    # Make an array of the days of which the LST is taken
    Dates = pd.date_range(Startdate, Enddate, freq = 'D')

    # Create Waitbar
    if Waitbar == 1:
        import pywapor.general.waitbar_console as WaitbarConsole
        total_amount = len(Dates)
        amount = 0
        WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    # Check the latitude and longitude and otherwise set lat or lon on greatest extent
    if latlim[0] < -90 or latlim[1] > 90:
        print('Latitude above 90N or below 90S is not possible. Value set to maximum')
        latlim[0] = np.max(latlim[0], -90)
        latlim[1] = np.min(latlim[1], 90)
    if lonlim[0] < -180 or lonlim[1] > 180:
        print('Longitude must be between 180E and 180W. Now value is set to maximum')
        lonlim[0] = np.max(lonlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)

    # Make directory for the MODIS NDVI data
    Dir = Dir.replace("/", os.sep)
    output_folder = os.path.join(Dir, 'MODIS', 'MOD11')
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Define which MODIS tiles are required
    TilesVertical, TilesHorizontal = Get_tiles_from_txt(latlim, lonlim) 
    
    # Pass variables to parallel function and run
    args = [output_folder, TilesVertical, TilesHorizontal,lonlim, latlim, username, password, hdf_library]
    all_dl_files = dict()
    for Date in Dates:
        dl_files = RetrieveData(Date, args)
        all_dl_files[Date.date()] = dl_files
        if Waitbar == 1:
            amount += 1
            WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    # Remove all .hdf files
    if remove_hdf == 1:
        try:
            os.chdir(output_folder)
            files = glob.glob("*.hdf")
            for f in files:
                os.remove(os.path.join(output_folder, f))
    
        except:
            pass

    return all_dl_files

def RetrieveData(Date, args):
    """
    This function retrieves MOD11 LST data for a given date from the
    https://e4ftl01.cr.usgs.gov/ server.

    Keyword arguments:
    Date -- 'yyyy-mm-dd'
    args -- A list of parameters defined in the DownloadData function.
    """
    
	# WAPOR modules
    import pywapor.general.processing_functions as PF
    
    # Argument
    [output_folder, TilesVertical, TilesHorizontal,lonlim, latlim, username, password, hdf_library] = args
    
    # Define output names
    LSTfileName = os.path.join(output_folder, 'LST_MOD11A1_K_daily_' + Date.strftime('%Y') + '.' + Date.strftime('%m') + '.' + Date.strftime('%d') + '.tif')
    TimefileName = os.path.join(output_folder, 'Time_MOD11A1_hour_daily_' + Date.strftime('%Y') + '.' + Date.strftime('%m') + '.' + Date.strftime('%d') + '.tif')    
    OnsangfileName = os.path.join(output_folder, 'Angle_MOD11A1_degrees_daily_' + Date.strftime('%Y') + '.' + Date.strftime('%m') + '.' + Date.strftime('%d') + '.tif')    

    if not (os.path.exists(LSTfileName) and os.path.exists(TimefileName) and os.path.exists(OnsangfileName)):

        Collect_data(TilesHorizontal, TilesVertical, Date, username, password, output_folder, hdf_library)
    
        # Define the output name of the collect data function
        name_collect = os.path.join(output_folder, 'Merged.tif')
        name_collect_time = os.path.join(output_folder, 'Merged_Time.tif')
        name_collect_obsang = os.path.join(output_folder, 'Merged_Obsang.tif')
        
        # Reproject the MODIS product to epsg_to
        epsg_to ='4326'
        name_reprojected = PF.reproject_MODIS(name_collect, epsg_to)
        name_reprojected_time = PF.reproject_MODIS(name_collect_time, epsg_to)
        name_reprojected_obsang = PF.reproject_MODIS(name_collect_obsang, epsg_to)
        
        # Clip the data to the users extend
        data, geo = PF.clip_data(name_reprojected, latlim, lonlim)
        data_time, geo = PF.clip_data(name_reprojected_time, latlim, lonlim)
        data_obsang, geo = PF.clip_data(name_reprojected_obsang, latlim, lonlim)
        
        # remove wrong values
        data[data==0.] = -9999
        data_time[data_time==25.5] = -9999
        
        # Save results as Gtiff 
        PF.Save_as_tiff(name=LSTfileName, data=data, geo=geo, projection='WGS84')
        PF.Save_as_tiff(name=TimefileName, data=data_time, geo=geo, projection='WGS84')
        PF.Save_as_tiff(name=OnsangfileName, data=data_obsang, geo=geo, projection='WGS84')
            
        # remove the side products
        os.remove(name_collect)
        os.remove(name_reprojected)
        os.remove(name_collect_time)
        os.remove(name_reprojected_time) 
        os.remove(name_collect_obsang)
        os.remove(name_reprojected_obsang)             

    return (LSTfileName, TimefileName, OnsangfileName)

def Collect_data(TilesHorizontal, TilesVertical, Date, username, password, output_folder, hdf_library):
    '''
    This function downloads all the needed MODIS tiles from https://e4ftl01.cr.usgs.gov/MOLT/MOD11A1.006/ as a hdf file.
    See documenation here: https://lpdaac.usgs.gov/products/mod11a1v006/
    
    Keywords arguments:
    TilesHorizontal -- [TileMin,TileMax] max and min horizontal tile number
    TilesVertical -- [TileMin,TileMax] max and min vertical tile number
    Date -- 'yyyy-mm-dd'
    output_folder -- 'C:/file/to/path/'
    '''

    # Make a new tile for the data
    sizeX = int((TilesHorizontal[1] - TilesHorizontal[0] + 1) * 1200)
    sizeY = int((TilesVertical[1] - TilesVertical[0] + 1) * 1200)
    DataTot = np.zeros((sizeY, sizeX))
    DataTot_Time = np.zeros((sizeY, sizeX))
    DataTot_ObsAng = np.zeros((sizeY, sizeX))
    
    # Create the Lat and Long of the MODIS tile in meters
    for Vertical in range(int(TilesVertical[0]), int(TilesVertical[1])+1):
        
        Distance = 4*231.65635826395834 # resolution of a MODIS pixel in meter
        countY=(TilesVertical[1] - TilesVertical[0] + 1) - (Vertical - TilesVertical[0])

        for Horizontal in range(int(TilesHorizontal[0]), int(TilesHorizontal[1]) + 1):
            countX=Horizontal - TilesHorizontal[0] + 1

            # Create the URL to the LST MODIS data
            url = 'https://e4ftl01.cr.usgs.gov/MOLT/MOD11A1.006/' + Date.strftime('%Y') + '.' + Date.strftime('%m') + '.' + Date.strftime('%d') + '/'

		    # Reset the begin parameters for downloading
            downloaded = 0
            N=0

			# Check the library given by user if the file is already there
            if hdf_library is not None:
                os.chdir(hdf_library)
                hdf_name = glob.glob("MOD11A1.A%s%03s.h%02dv%02d.*" %(Date.strftime('%Y'), Date.strftime('%j'), Horizontal, Vertical))

                if len(hdf_name) == 1:
                    hdf_file = os.path.join(hdf_library, hdf_name[0])

                    if os.path.exists(hdf_file):
                        downloaded = 1
                        file_name = hdf_file
                        
            # If the file is not existing, download the data
            if not downloaded == 1:

                # Get files on FTP server
                if sys.version_info[0] == 3:
                    f = urllib.request.urlopen(url)

                if sys.version_info[0] == 2:
                    f = urllib2.urlopen(url)

                # Sum all the files on the server
                soup = BeautifulSoup(f, "lxml")
                for i in soup.findAll('a', attrs = {'href': re.compile('(?i)(hdf)$')}):

                    # Find the file with the wanted tile number
                    Vfile=str(i)[30:32]
                    Hfile=str(i)[27:29]
                    
                    if int(Vfile) is int(Vertical) and int(Hfile) is int(Horizontal):

                        # Define the whole url name
                        if sys.version_info[0] == 3:
                            full_url = urllib.parse.urljoin(url, i['href'])

                        if sys.version_info[0] == 2:
                            full_url = urlparse.urljoin(url, i['href'])

                        # if not downloaded try to download file
                        while downloaded == 0:

                            try: # open http and download whole .hdf
                                nameDownload = full_url
                                file_name = os.path.join(output_folder,nameDownload.split('/')[-1])
                                if os.path.isfile(file_name):
                                    print("file ", file_name, " already exists")
                                    downloaded = 1
                                else:
                                    x = requests.get(nameDownload, allow_redirects = False)
                                    try:
                                        y = requests.get(x.headers['location'], auth = (username, password))
                                    except:
                                        # from requests.packages.urllib3.exceptions import InsecureRequestWarning
                                        # requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

                                        y = requests.get(x.headers['location'], auth = (username, password), verify = False)
                                        
                                    z = open(file_name, 'wb')
                                    z.write(y.content)
                                    z.close()
                                    statinfo = os.stat(file_name)
                                    # Say that download was succesfull
                                    if int(statinfo.st_size) > 10000:
                                         downloaded = 1

                            # If download was not succesfull
                            except:

                                # Try another time
                                N = N + 1

    				    # Stop trying after 10 times
                        if N == 10:
                            print('Data from ' + Date.strftime('%Y-%m-%d') + ' is not available')
                            downloaded = 1
                            
            try:
                # Open .hdf only band with LST and collect all tiles to one array
                dataset = gdal.Open(file_name)
                sdsdict = dataset.GetMetadata('SUBDATASETS')
                sdslist = [sdsdict[k] for k in sdsdict.keys() if (('SUBDATASET_1_NAME') in k or ('SUBDATASET_3_NAME') in k or ('SUBDATASET_4_NAME') in k)]
                sds = []
                sds_time = []
                sds_obsang = []
                                
                full_layer = [i for i in sdslist if 'LST_Day_1km' in i]
                full_layer_time = [i for i in sdslist if 'Day_view_time' in i]
                full_layer_obsang = [i for i in sdslist if 'Day_view_angl' in i]
                idx = sdslist.index(full_layer[0])
                idx_time = sdslist.index(full_layer_time[0])
                idx_obsang = sdslist.index(full_layer_obsang[0])
                sds.append(gdal.Open(sdslist[idx]))
                sds_time.append(gdal.Open(sdslist[idx_time]))     
                sds_obsang.append(gdal.Open(sdslist[idx_obsang]))                  
                if Horizontal == TilesHorizontal[0] and Vertical == TilesVertical[0]:
                    geo_t = sds[idx].GetGeoTransform()

                    # get the projection value
                    proj = sds[idx].GetProjection()

                # LST_Day_1km	Daytime Land Surface Temperature	Kelvin	16-bit unsigned integer	0	N/A	7500 to 65535	0.02	N/A
                data = sds[0].ReadAsArray()
                # Day_view_time	Local time of day observation	Hours	8-bit unsigned integer	255	N/A	0 to 240	0.1	N/A
                data_time = sds_time[0].ReadAsArray()
                # Day_view_angl	View zenith angle of day observation	Degree	8-bit unsigned integer	255	N/A	0 to 130	N/A	-65
                data_obsang = sds_obsang[0].ReadAsArray()

                
                countYdata = (TilesVertical[1] - TilesVertical[0] + 2) - countY
                DataTot[int((countYdata - 1) * 1200):int(countYdata * 1200), int((countX - 1) * 1200):int(countX * 1200)]=data * 0.02
                DataTot_Time[int((countYdata - 1) * 1200):int(countYdata * 1200), int((countX - 1) * 1200):int(countX * 1200)]=data_time * 0.1
                DataTot_ObsAng[int((countYdata - 1) * 1200):int(countYdata * 1200), int((countX - 1) * 1200):int(countX * 1200)]= data_obsang
                del data, data_time, data_obsang

            # if the tile not exists or cannot be opened, create a nan array with the right projection
            except:
                if Horizontal==TilesHorizontal[0] and Vertical==TilesVertical[0]:
                     x1 = (TilesHorizontal[0] - 19) * 1200 * Distance
                     x4 = (TilesVertical[0] - 9) * 1200 * -1 * Distance
                     geo = 	[x1, Distance, 0.0, x4, 0.0, -Distance]
                     geo_t=tuple(geo)

                proj='PROJCS["unnamed",GEOGCS["Unknown datum based upon the custom spheroid",DATUM["Not specified (based on custom spheroid)",SPHEROID["Custom spheroid",6371007.181,0]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],PROJECTION["Sinusoidal"],PARAMETER["longitude_of_center",0],PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["Meter",1]]'
                data=np.ones((1200,1200)) * (-9999/0.02)
                data_time=np.ones((1200,1200)) * (-9999/0.1)
                data_obsang=np.ones((1200,1200)) * 255
                countYdata=(TilesVertical[1] - TilesVertical[0] + 2) - countY
                DataTot[(countYdata - 1) * 1200:countYdata * 1200,(countX - 1) * 1200:countX * 4800] = data * 0.02
                DataTot_Time[(countYdata - 1) * 1200:countYdata * 1200,(countX - 1) * 1200:countX * 4800] = data_time * 0.1
                DataTot_ObsAng[(countYdata - 1) * 1200:countYdata * 1200,(countX - 1) * 1200:countX * 4800] = data_obsang
                
    # Set limits    
    DataTot_Time[np.logical_and(DataTot_Time < 0., DataTot_Time>24.)] = -9999
    DataTot[DataTot < 1] = -9999   
    DataTot_ObsAng[DataTot_ObsAng<255] = DataTot_ObsAng[DataTot_ObsAng<255] - 65
    DataTot_ObsAng[DataTot_ObsAng==255] = -9999
    DataTot_ObsAng[DataTot_ObsAng>255] = DataTot_ObsAng[DataTot_ObsAng>255] - 256
    
    # Make geotiff file of the LST data
    name_out_LST = os.path.join(output_folder, 'Merged.tif')
    driver = gdal.GetDriverByName("GTiff")
    dst_ds = driver.Create(name_out_LST, DataTot.shape[1], DataTot.shape[0], 1, gdal.GDT_Float32, ['COMPRESS=LZW'])
    try:
         dst_ds.SetProjection(proj)
    except:
        proj='PROJCS["unnamed",GEOGCS["Unknown datum based upon the custom spheroid",DATUM["Not specified (based on custom spheroid)",SPHEROID["Custom spheroid",6371007.181,0]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],PROJECTION["Sinusoidal"],PARAMETER["longitude_of_center",0],PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["Meter",1]]'
        x1 = (TilesHorizontal[0] - 18) * 1200 * Distance
        x4 = (TilesVertical[0] - 9) * 1200 * -1 * Distance
        geo = [x1, Distance, 0.0, x4, 0.0, -Distance]
        geo_t = tuple(geo)
        dst_ds.SetProjection(proj)

    dst_ds.GetRasterBand(1).SetNoDataValue(-9999)
    dst_ds.SetGeoTransform(geo_t)
    dst_ds.GetRasterBand(1).WriteArray(DataTot)
    dst_ds = None
    sds = None

    # Make geotiff file of the time data
    name_out_time = os.path.join(output_folder, 'Merged_Time.tif')
    driver = gdal.GetDriverByName("GTiff")
    dst_ds = driver.Create(name_out_time, DataTot_Time.shape[1], DataTot_Time.shape[0], 1, gdal.GDT_Float32, ['COMPRESS=LZW'])
    try:
         dst_ds.SetProjection(proj)
    except:
        proj='PROJCS["unnamed",GEOGCS["Unknown datum based upon the custom spheroid",DATUM["Not specified (based on custom spheroid)",SPHEROID["Custom spheroid",6371007.181,0]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],PROJECTION["Sinusoidal"],PARAMETER["longitude_of_center",0],PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["Meter",1]]'
        x1 = (TilesHorizontal[0] - 18) * 1200 * Distance
        x4 = (TilesVertical[0] - 9) * 1200 * -1 * Distance
        geo = [x1, Distance, 0.0, x4, 0.0, -Distance]
        geo_t = tuple(geo)
        dst_ds.SetProjection(proj)

    dst_ds.GetRasterBand(1).SetNoDataValue(-9999)
    dst_ds.SetGeoTransform(geo_t)
    dst_ds.GetRasterBand(1).WriteArray(DataTot_Time)
    dst_ds = None
    sds = None    

    # Make geotiff file of the time data
    name_out_obsang = os.path.join(output_folder, 'Merged_Obsang.tif')
    driver = gdal.GetDriverByName("GTiff")
    dst_ds = driver.Create(name_out_obsang, DataTot_ObsAng.shape[1], DataTot_ObsAng.shape[0], 1, gdal.GDT_Float32, ['COMPRESS=LZW'])
    try:
         dst_ds.SetProjection(proj)
    except:
        proj='PROJCS["unnamed",GEOGCS["Unknown datum based upon the custom spheroid",DATUM["Not specified (based on custom spheroid)",SPHEROID["Custom spheroid",6371007.181,0]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],PROJECTION["Sinusoidal"],PARAMETER["longitude_of_center",0],PARAMETER["false_easting",0],PARAMETER["false_northing",0],UNIT["Meter",1]]'
        x1 = (TilesHorizontal[0] - 18) * 1200 * Distance
        x4 = (TilesVertical[0] - 9) * 1200 * -1 * Distance
        geo = [x1, Distance, 0.0, x4, 0.0, -Distance]
        geo_t = tuple(geo)
        dst_ds.SetProjection(proj)

    dst_ds.GetRasterBand(1).SetNoDataValue(-9999)
    dst_ds.SetGeoTransform(geo_t)
    dst_ds.GetRasterBand(1).WriteArray(DataTot_ObsAng)
    dst_ds = None
    sds = None    
    
    return()
    
def Get_tiles_from_txt(latlim, lonlim):
    '''
    Opens a text file storing the extents of every MODIS tile

    Keywords arguments:
    tiletext2 -- 'C:/file/to/path/' to path of the txt file with all the MODIS tiles extents
    lonlim1 -- [ymin, ymax] (longitude limits of the chunk or whole image)
    latlim1 -- [ymin, ymax] (latitude limits of the chunk or whole image)
    '''

    # Get the list (txt file from the internet) which includes the lat and lon information of the MODIS tiles
    path = os.path.abspath(os.path.dirname(__file__))
    file_nametext = os.path.join(path, "sn_gring_10deg.txt")

    # Open text file with tiles which is downloaded before
    tiletext=np.genfromtxt(file_nametext,skip_header=7,skip_footer=1,usecols=(0,1,2,3,4,5,6,7,8,9))
    tiletext2=tiletext[tiletext[:,2]>=-900,:]

    # This function converts the values in the text file into horizontal and vertical number of the tiles which must be downloaded to cover the extent defined by the user
    TilesVertical, TilesHorizontal = Tiles_to_download(tiletext2=tiletext2,lonlim1=lonlim,latlim1=latlim)

    return(TilesVertical, TilesHorizontal)

def Tiles_to_download(tiletext2,lonlim1,latlim1):
    '''
    Defines the MODIS tiles that must be downloaded in order to cover the latitude and longitude limits

    Keywords arguments:
    tiletext2 -- 'C:/file/to/path/' to path of the txt file with all the MODIS tiles extents
    lonlim1 -- [ymin, ymax] (longitude limits of the chunk or whole image)
    latlim1 -- [ymin, ymax] (latitude limits of the chunk or whole image)
    '''
    # calculate min and max longitude and latitude
    # lat down    lat up      lon left     lon right
    tiletextExtremes = np.empty([len(tiletext2),6])
    tiletextExtremes[:,0] = tiletext2[:,0]
    tiletextExtremes[:,1] = tiletext2[:,1]
    tiletextExtremes[:,2] = np.minimum(tiletext2[:,3], tiletext2[:,9])
    tiletextExtremes[:,3] = np.maximum(tiletext2[:,5], tiletext2[:,7])
    tiletextExtremes[:,4] = np.minimum(tiletext2[:,2], tiletext2[:,4])
    tiletextExtremes[:,5] = np.maximum(tiletext2[:,6], tiletext2[:,8])

    # Define the upper left tile
    latlimtiles1UL = tiletextExtremes[np.logical_and(tiletextExtremes[:,2] <= latlim1[1], tiletextExtremes[:,3] >= latlim1[1])]
    latlimtilesUL = latlimtiles1UL[np.logical_and(latlimtiles1UL[:,4] <= lonlim1[0], latlimtiles1UL[:,5] >= lonlim1[0])]

    # Define the lower right tile
    latlimtiles1LR = tiletextExtremes[np.logical_and(tiletextExtremes[:,2] <= latlim1[0], tiletextExtremes[:,3] >= latlim1[0])]
    latlimtilesLR = latlimtiles1LR[np.logical_and(latlimtiles1LR[:,4]<=lonlim1[1],latlimtiles1LR[:,5]>=lonlim1[1])]

    # Define the total tile
    TotalTiles = np.vstack([latlimtilesUL, latlimtilesLR])

    # Find the minimum horizontal and vertical tile value and the maximum horizontal and vertical tile value
    TilesVertical = [TotalTiles[:,0].min(), TotalTiles[:,0].max()]
    TilesHorizontal = [TotalTiles[:,1].min(), TotalTiles[:,1].max()]
    
    return(TilesVertical, TilesHorizontal)    



# -*- coding: utf-8 -*-
"""
"""
import os
from osgeo import osr
from osgeo import gdal
import gzip
import zipfile
import numpy as np
from scipy.interpolate import NearestNDInterpolator, LinearNDInterpolator

def apply_mask(a, indices, axis):
    """
    https://stackoverflow.com/questions/15469302/numpy-3d-to-2d-transformation-based-on-2d-mask-array
    # replace with np.take(....)?
    """
    magic_index = [np.arange(i) for i in indices.shape]
    magic_index = np.ix_(*magic_index)
    magic_index = magic_index[:axis] + (indices,) + magic_index[axis:]
    return a[magic_index]

def reproj_file(file, template, method):
    ds = reproject_dataset_example(file, template, method = method)
    array = open_as_array(ds)
    return array

def combine_dicts(dicts):
    new_dict = dict()
    for d in dicts:
        for key, value in d.items():
            if key in new_dict.keys():
                new_dict[key].append(value)
            else:
                new_dict[key] = [value]
    return new_dict

def get_geoinfo(template_file):
    if isinstance(template_file, str):
        ds = gdal.Open(template_file)
    elif isinstance(template_file, gdal.Dataset):
        ds = template_file
    geo_ex = ds.GetGeoTransform()
    proj_ex = ds.GetProjection()
    size_x_ex = ds.RasterXSize
    size_y_ex = ds.RasterYSize
    return (geo_ex, proj_ex, size_x_ex, size_y_ex)

def Extract_Data_gz(zip_filename, outfilename):
    """
    This function extract the zip files

    Keyword Arguments:
    zip_filename -- name, name of the file that must be unzipped
    outfilename -- Dir, directory where the unzipped data must be
                           stored
    """

    with gzip.GzipFile(zip_filename, 'rb') as zf:
        file_content = zf.read()
        save_file_content = open(outfilename, 'wb')
        save_file_content.write(file_content)
    save_file_content.close()
    zf.close()
    os.remove(zip_filename)
    
def Save_as_MEM(data='', geo='', projection=''):
    """
    This function save the array as a memory file

    Keyword arguments:
    data -- [array], dataset of the geotiff
    geo -- [minimum lon, pixelsize, rotation, maximum lat, rotation,
            pixelsize], (geospatial dataset)
    projection -- interger, the EPSG code
    """
    # save as a geotiff
    driver = gdal.GetDriverByName("MEM")
    dst_ds = driver.Create('', int(data.shape[1]), int(data.shape[0]), 1,
                           gdal.GDT_Float32)
    srse = osr.SpatialReference()
    if projection == '':
        srse.SetWellKnownGeogCS("WGS84")

    else:
        try:
            if not srse.SetWellKnownGeogCS(projection) == 6:
                srse.SetWellKnownGeogCS(projection)
            else:
                try:
                    srse.ImportFromEPSG(int(projection))
                except:
                    srse.ImportFromWkt(projection)
        except:
            try:
                srse.ImportFromEPSG(int(projection))
            except:
                srse.ImportFromWkt(projection)
    dst_ds.SetProjection(srse.ExportToWkt())
    dst_ds.GetRasterBand(1).SetNoDataValue(-9999)
    dst_ds.SetGeoTransform(geo)
    dst_ds.GetRasterBand(1).WriteArray(data)
    return(dst_ds)   
    
def Save_as_tiff(name='', data='', geo='', projection=''):
    """
    This function save the array as a geotiff

    Keyword arguments:
    name -- string, directory name
    data -- [array], dataset of the geotiff
    geo --  [minimum lon, pixelsize, rotation, maximum lat, rotation,
            pixelsize], (geospatial dataset)
    projection -- integer, the EPSG code
    """
    # Change no data values
    data[np.isnan(data)] = -9999

    if not os.path.exists(os.path.split(name)[0]):
        os.makedirs(os.path.split(name)[0])
    
    # save as a geotiff
    driver = gdal.GetDriverByName("GTiff")
    dst_ds = driver.Create(name, int(data.shape[1]), int(data.shape[0]), 1,
                           gdal.GDT_Float32, ['COMPRESS=LZW', 'PREDICTOR=3'])
    srse = osr.SpatialReference()
    if projection == '':
        srse.SetWellKnownGeogCS("WGS84")

    # Set the projection, which can be an EPSG code or a well known GeogCS
    else:
        try:
            if not srse.SetWellKnownGeogCS(projection) == 6:
                srse.SetWellKnownGeogCS(projection)
            else:
                try:
                    srse.ImportFromEPSG(int(projection))
                except:
                    srse.ImportFromWkt(projection)
        except:
            try:
                srse.ImportFromEPSG(int(projection))
            except:
                srse.ImportFromWkt(projection)
                
    # Save the tiff file
    dst_ds.SetProjection(srse.ExportToWkt())
    dst_ds.GetRasterBand(1).SetNoDataValue(-9999)
    dst_ds.SetGeoTransform(geo)
    dst_ds.GetRasterBand(1).WriteArray(data)
    dst_ds = None
    
    return()
    
def reproject_MODIS(input_name, epsg_to):
    '''
    Reproject the merged data file by using gdalwarp. The input projection must be the MODIS projection.
    The output projection can be defined by the user.

    Keywords arguments:
    input_name -- 'C:/file/to/path/file.tif'
        string that defines the input tiff file
    epsg_to -- integer
        The EPSG code of the output dataset
    '''
    
    # Define the output name
    name_out = ''.join(input_name.split(".")[:-1]) + '_reprojected.tif'
   
    src_ds = gdal.Open(input_name)
    
    # Define target SRS
    dst_srs = osr.SpatialReference()
    dst_srs.ImportFromEPSG(int(epsg_to))
    dst_wkt = dst_srs.ExportToWkt()
    
    error_threshold = 0.125  # error threshold --> use same value as in gdalwarp
    resampling = gdal.GRA_NearestNeighbour
    
    # Call AutoCreateWarpedVRT() to fetch default values for target raster dimensions and geotransform
    tmp_ds = gdal.AutoCreateWarpedVRT( src_ds,
                                   None, # src_wkt : left to default value --> will use the one from source
                                   dst_wkt,
                                   resampling,
                                   error_threshold )
    dst_ds = gdal.GetDriverByName('GTiff').CreateCopy(name_out, tmp_ds)
    dst_ds = None 

    return(name_out)
    
def clip_data(input_file, latlim, lonlim):
    """
    Clip the data to the defined extend of the user (latlim, lonlim)

    Keyword Arguments:
    input_file -- output data, output of the clipped dataset
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    """
    try:
        if input_file.split('.')[-1] == 'tif':
            dest_in = gdal.Open(input_file)
        else:
            dest_in = input_file
    except:
        dest_in = input_file

    # Open Array
    data_in = dest_in.GetRasterBand(1).ReadAsArray()

    # Define the array that must remain
    Geo_in = dest_in.GetGeoTransform()
    Geo_in = list(Geo_in)
    Start_x = np.max([int(np.floor(((lonlim[0]) - Geo_in[0])/ Geo_in[1])),0])
    End_x = np.min([int(np.ceil(((lonlim[1]) - Geo_in[0])/ Geo_in[1])),int(dest_in.RasterXSize)])

    Start_y = np.max([int(np.floor((Geo_in[3] - latlim[1])/ -Geo_in[5])),0])
    End_y = np.min([int(np.ceil(((latlim[0]) - Geo_in[3])/Geo_in[5])), int(dest_in.RasterYSize)])

    #Create new GeoTransform
    Geo_in[0] = Geo_in[0] + Start_x * Geo_in[1]
    Geo_in[3] = Geo_in[3] + Start_y * Geo_in[5]
    Geo_out = tuple(Geo_in)

    data = np.zeros([End_y - Start_y, End_x - Start_x])

    data = data_in[Start_y:End_y,Start_x:End_x]
    dest_in = None

    return(data, Geo_out)    
    
    
def Extract_Data(input_file, output_folder):
    """
    This function extract the zip files

    Keyword Arguments:
    output_file -- name, name of the file that must be unzipped
    output_folder -- Dir, directory where the unzipped data must be
                           stored
    """
    # extract the data
    z = zipfile.ZipFile(input_file, 'r')
    z.extractall(output_folder)
    z.close()    

def reproj_ds(source_file, example_file, ndv = -9999, override_dtype = True):
    
    source_ds = gdal.Open(source_file)

    example_ds = gdal.Open(example_file)
    example_xsize = example_ds.RasterXSize
    example_ysize = example_ds.RasterYSize

    if override_dtype:
        example_dtype = example_ds.GetRasterBand(1).DataType
    else:
        example_dtype = source_ds.GetRasterBand(1).DataType

    mem_drv = gdal.GetDriverByName('MEM')
    dest_ds = mem_drv.Create('', example_xsize, example_ysize, 1, example_dtype)
    dest_ds.SetGeoTransform(example_ds.GetGeoTransform())
    dest_ds.SetProjection(example_ds.GetProjection())
    dest_ds.GetRasterBand(1).Fill(ndv)
    dest_ds.GetRasterBand(1).SetNoDataValue(ndv)

    gdal.ReprojectImage(source_ds,
                        dest_ds,
                        source_ds.GetProjection(),
                        dest_ds.GetProjection(),
                        gdal.GRA_NearestNeighbour)

    return dest_ds

def reproject_dataset_example(dataset, dataset_example, method=1):
    """
    A sample function to reproject and resample a GDAL dataset from within
    Python. The user can define the wanted projection and shape by defining an example dataset.

    Keywords arguments:
    dataset -- 'C:/file/to/path/file.tif' or a gdal file (gdal.Open(filename))
        string that defines the input tiff file or gdal file
    dataset_example -- 'C:/file/to/path/file.tif' or a gdal file (gdal.Open(filename))
        string that defines the input tiff file or gdal file
    method -- 1,2,3,4 default = 1
        1 = Nearest Neighbour, 2 = Bilinear, 3 = lanzcos, 4 = average
    """
    # open dataset that must be transformed
    try:
        if os.path.splitext(dataset)[-1] == '.tif':
            g = gdal.Open(dataset)
        else:
            g = dataset
    except:
            g = dataset
    epsg_from = Get_epsg(g)

    #exceptions
    if epsg_from == 9001:
        epsg_from = 5070

    # open dataset that is used for transforming the dataset
    try:
        gland = gdal.Open(dataset_example)
        epsg_to = Get_epsg(gland)
    except:
        gland = dataset_example
        epsg_to = Get_epsg(gland)

    # Set the EPSG codes
    osng = osr.SpatialReference()
    osng.ImportFromEPSG(epsg_to)
    wgs84 = osr.SpatialReference()
    wgs84.ImportFromEPSG(epsg_from)

    # Get shape and geo transform from example
    geo_land = gland.GetGeoTransform()
    col=gland.RasterXSize
    rows=gland.RasterYSize

    # Create new raster
    mem_drv = gdal.GetDriverByName('MEM')
    dest = mem_drv.Create('', col, rows, 1, gdal.GDT_Float32)
    dest.SetGeoTransform(geo_land)
    dest.SetProjection(osng.ExportToWkt())

    # https://gis.stackexchange.com/questions/158503/9999-no-data-value-becomes-0-when-writing-array-to-gdal-memory-file
    ndv = g.GetRasterBand(1).GetNoDataValue()
    band = dest.GetRasterBand(1)
    band.SetNoDataValue(ndv)
    band.Fill(ndv, 0.0)

    # Perform the projection/resampling
    if method == 1:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_NearestNeighbour)
    if method == 2:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Bilinear)
    if method == 3:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Lanczos)
    if method == 4:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Average)
    if method == 5:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Cubic)
    if method == 6:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_CubicSpline)
    if method == 7:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Mode)
    if method == 8:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Max)
    if method == 9:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Min)
    if method == 10:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Med)
    if method == 11:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Q1)
    if method == 12:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Q3)
    if method == 13:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Sum)
        
    return(dest)

def Get_epsg(g):
    """
    This function reads the projection of a GEOGCS file or tiff file

    Keyword arguments:
    g -- string
        Filename to the file that must be read
    extension -- tiff or GEOGCS
        Define the extension of the dataset (default is tiff)
    """
    try:
        # Get info of the dataset that is used for transforming
        g_proj = g.GetProjection()
        Projection=g_proj.split('EPSG","')
        epsg_to=int((str(Projection[-1]).split(']')[0])[0:-1])            
    except:
        epsg_to=4326

    return(epsg_to)

def Show_tif(image_file, Limits = None, Color = None):
    """
    This function plot a tiff array in the console

    Keyword arguments:
    image_file -- string
        Filename to the file that must be shown
    Limits -- [min, max] (Default = min and max image)
        User can define the limits of the colorbar
    Color -- string (Default = "viridis")
        User can define the wanted colormap, all options are listed here:
        https://matplotlib.org/examples/color/colormaps_reference.html
    """    
    dest = gdal.Open(image_file)
    Array = dest.GetRasterBand(1).ReadAsArray()
    Array[Array==-9999] = np.nan
    if Limits == None:
        Limits = [np.nanmin(Array), np.nanmax(Array)]
    
    if Color == None:
        Color = "viridis"
    
    import matplotlib.pyplot as plt
    
    plt.imshow(Array, cmap = Color, vmin=Limits[0], vmax=Limits[1])
    plt.colorbar()
    plt.show()
    
    return()

def open_as_array(input):
    if isinstance(input, str):
        ds = gdal.Open(input)
    elif isinstance(input, gdal.Dataset):
        ds = input
    array = ds.GetRasterBand(1).ReadAsArray().astype(float)
    ndv = ds.GetRasterBand(1).GetNoDataValue()
    array[array == ndv] = np.nan
    return array  

def Open_tiff_array(filename='', band=''):
    """
    Opening a tiff array.

    Keyword Arguments:
    filename -- 'C:/file/to/path/file.tif' or a gdal file (gdal.Open(filename))
        string that defines the input tiff file or gdal file
    band -- integer
        Defines the band of the tiff that must be opened.
    """
    f = gdal.Open(filename)
    if f is None:
        print('%s does not exists' %filename)
    else:
        if band == '':
            band = 1
        Data = f.GetRasterBand(band).ReadAsArray()
    return(Data)

def Open_array_info(filename=''):
    """
    Opening a tiff info, for example size of array, projection and transform matrix.

    Keyword Arguments:
    filename -- 'C:/file/to/path/file.tif' or a gdal file (gdal.Open(filename))
        string that defines the input tiff file or gdal file

    """
    try:
        if filename.split('.')[-1] == 'tif':
            f = gdal.Open(r"%s" %filename)
        else:
            f = filename
    except:
            f = filename       
    try:
        geo_out = f.GetGeoTransform()
        proj = f.GetProjection()
        size_X = f.RasterXSize
        size_Y = f.RasterYSize
        f = None
    except:
        print('%s does not exists' %filename)
        
    return(geo_out, proj, size_X, size_Y)

def gap_filling(dataset, NoDataValue, method = 1):
    """
    This function fills the no data gaps in a numpy array

    Keyword arguments:
    dataset -- 'C:/'  path to the source data (dataset that must be filled)
    NoDataValue -- Value that must be filled
    """
    try:
        if dataset.split('.')[-1] == 'tif':
            # Open the numpy array
            data = Open_tiff_array(dataset)
            Save_as_tiff_bool = 1
        else:
            data = dataset
            Save_as_tiff_bool = 0
    except:
        data = dataset
        Save_as_tiff_bool = 0

    # fill the no data values
    if NoDataValue == np.nan:
        mask = ~(np.isnan(data))
    else:
        mask = ~(data==NoDataValue)
    xx, yy = np.meshgrid(np.arange(data.shape[1]), np.arange(data.shape[0]))
    xym = np.vstack( (np.ravel(xx[mask]), np.ravel(yy[mask])) ).T
    data0 = np.ravel( data[:,:][mask] )

    if method == 1:
        interp0 = NearestNDInterpolator( xym, data0 )
        data_end = interp0(np.ravel(xx), np.ravel(yy)).reshape( xx.shape )

    if method == 2:
        interp0 = LinearNDInterpolator( xym, data0 )
        data_end = interp0(np.ravel(xx), np.ravel(yy)).reshape( xx.shape )

    if Save_as_tiff_bool == 1:
        EndProduct=dataset[:-4] + '_GF.tif'

        # collect the geoinformation
        geo_out, proj, size_X, size_Y = Open_array_info(dataset)

        # Save the filled array as geotiff
        Save_as_tiff(name=EndProduct, data=data_end, geo=geo_out, projection=proj)

    else:
        EndProduct = data_end

    return (EndProduct)  
    
def calc_dlat_dlon(geo_out, size_X, size_Y):
    """
    This functions calculated the distance between each pixel in meter.

    Parameters
    ----------
    geo_out: array
        geo transform function of the array
    size_X: int
        size of the X axis
    size_Y: int
        size of the Y axis

    Returns
    -------
    dlat: array
        Array containing the vertical distance between each pixel in meters
    dlon: array
        Array containing the horizontal distance between each pixel in meters
    """

    # Create the lat/lon rasters
    lon = np.arange(size_X + 1)*geo_out[1]+geo_out[0] - 0.5 * geo_out[1]
    lat = np.arange(size_Y + 1)*geo_out[5]+geo_out[3] - 0.5 * geo_out[5]

    dlat_2d = np.array([lat,]*int(np.size(lon,0))).transpose()
    dlon_2d =  np.array([lon,]*int(np.size(lat,0)))

    # Radius of the earth in meters
    R_earth = 6371000

    # Calculate the lat and lon in radians
    lonRad = dlon_2d * np.pi/180
    latRad = dlat_2d * np.pi/180

    # Calculate the difference in lat and lon
    lonRad_dif = abs(lonRad[:,1:] - lonRad[:,:-1])
    latRad_dif = abs(latRad[:-1] - latRad[1:])

    # Calculate the distance between the upper and lower pixel edge
    a = np.sin(latRad_dif[:,:-1]/2) * np.sin(latRad_dif[:,:-1]/2)
    clat = 2 * np.arctan2(np.sqrt(a), np.sqrt(1-a))
    dlat = R_earth * clat

    # Calculate the distance between the eastern and western pixel edge
    b = np.cos(latRad[1:,:-1]) * np.cos(latRad[:-1,:-1]) * np.sin(lonRad_dif[:-1,:]/2) * np.sin(lonRad_dif[:-1,:]/2)
    clon = 2 * np.arctan2(np.sqrt(b), np.sqrt(1-b))
    dlon = R_earth * clon

    return(dlat, dlon)  
    



# -*- coding: utf-8 -*-
"""
"""
import os
from datetime import datetime as dat
from osgeo import gdal
import pandas as pd
import numpy as np
import requests
import pywapor
import pywapor.general.processing_functions as PF
import pywapor.collect as c
import pywapor.general as g
import json
import xarray as xr
import glob
import warnings
import shutil

def main(project_folder, startdate, enddate, latlim, lonlim, level = "level_1", 
        diagnostics = None, composite_length = "DEKAD"):

    print("\n#################")
    print("## PRE_ET_LOOK ##")
    print("#################")

#%%
    # Disable Warnings
    warnings.filterwarnings('ignore')
    all_files = dict()

    sdate = dat.strptime(startdate, "%Y-%m-%d").date()
    edate = dat.strptime(enddate, "%Y-%m-%d").date()
    epochs_info = create_dates(sdate, edate, composite_length)

    #### Check if selected sources and dates are valid
    levels = g.variables.get_source_level_selections()

    if isinstance(level, dict):
        level_name = "custom"
        if "name" in level.keys():
            level_name = level.pop("name")
        levels[level_name] = level
        level = level_name

    source_selection = levels[level]
    succes = g.tests.check_source_selection(source_selection, sdate, edate)[1]
    assert succes, "invalid source_selection"

    raw_folder = os.path.join(project_folder, "RAW")
    level_folder = os.path.join(project_folder, level)
    temp_folder = os.path.join(project_folder, "temporary")
    if diagnostics:
        diagnostics["folder"] = os.path.join(level_folder, "graphs")

    dl_args = (raw_folder, latlim, lonlim, startdate, enddate)

    unraw_file_templates = unraw_filepaths(epochs_info[1], level_folder, "{var}")

    #### NDVI ####
    print(f"\n#### NDVI ####")
    raw_ndvi_files = list()
    # Order is important! PROBV gets priority over MOD13, and MOD13 over MYD13.
    if "PROBAV" in source_selection["NDVI"]:
        raw_ndvi_files.append(c.PROBAV.PROBAV_S5(*dl_args)[0])
        # raw_ndvi_files.append(glob.glob(os.path.join(project_folder, "RAW/PROBAV/NDVI/*.tif")))
    if "MOD13" in source_selection["NDVI"]:
        raw_ndvi_files.append(c.MOD13.NDVI(*dl_args))
    if "MYD13" in source_selection["NDVI"]:
        raw_ndvi_files.append(c.MYD13.NDVI(*dl_args))

    example_fh, example_ds, example_geoinfo = select_template(raw_ndvi_files)

    cmeta = {
        "composite_type": "mean",
        "temporal_interp": "linear",
        "spatial_interp": "nearest",
        "var_name": "ndvi",
        "var_unit": "-",
    }
    ds = g.compositer.main(cmeta, raw_ndvi_files, epochs_info, temp_folder, example_ds, 
                                diagnostics = diagnostics)

    all_files["ndvi"] = unraw_all(cmeta["var_name"], ds, unraw_file_templates, example_geoinfo)

#%%

    #### ALBEDO ####
    print(f"\n#### ALBEDO ####")
    raw_albedo_files = list()
    # Order is important! PROBV gets priority over MDC43.
    if "PROBAV" in source_selection["ALBEDO"]:
        raw_albedo_files.append(c.PROBAV.PROBAV_S5(*dl_args)[1])
        # raw_albedo_files.append(glob.glob(r"/Volumes/Data/FAO/WaPOR_vs_pyWaPOR/pyWAPOR_long_test/RAW/PROBAV/Albedo/*.tif"))
    if "MDC43" in source_selection["ALBEDO"]:
        raw_albedo_files.append(c.MCD43.ALBEDO(*dl_args))

    cmeta = {
        "composite_type": "mean",
        "temporal_interp": "linear",
        "spatial_interp": "nearest",
        "var_name": "r0",
        "var_unit": "-",
    }
    ds = g.compositer.main(cmeta, raw_albedo_files, epochs_info, temp_folder, example_ds, diagnostics = diagnostics)
    all_files["r0"] = unraw_all(cmeta["var_name"], ds, unraw_file_templates, example_geoinfo)

#%%
    #### PRECIPITATION ####
    print(f"\n#### PRECIPITATION ####")
    raw_precip_files = list()
    if "CHIRPS" in source_selection["PRECIPITATION"]:
        raw_precip_files.append(c.CHIRPS.daily(*dl_args))

    cmeta = {
        "composite_type": "mean",
        "temporal_interp": "linear",
        "spatial_interp": "nearest",
        "var_name": "P_24",
        "var_unit": "mm/day",
    }
    ds = g.compositer.main(cmeta, raw_precip_files, epochs_info, temp_folder, example_ds, diagnostics = diagnostics)
    all_files["P_24"] = unraw_all(cmeta["var_name"], ds, unraw_file_templates, example_geoinfo)

#%%
    #### DEM ####
    print(f"\n#### DEM ####")
    if "SRTM" in source_selection["DEM"]:
        raw_dem_file = c.SRTM.DEM(*dl_args[:3])
    
    dem_file = unraw_filepaths(None, level_folder, "z", static = True)[0]

    cmeta = {
        "composite_type": None,
        "temporal_interp": None,
        "spatial_interp": "linear",
        "var_name": "z",
        "var_unit": "m",
    }
    ds = g.compositer.main(cmeta, [[raw_dem_file]], None, temp_folder, example_ds)
    all_files["z"] = unraw_all(cmeta["var_name"], ds, [dem_file], example_geoinfo)

    #### SLOPE ASPECT ####
    print(f"\n#### SLOPE ASPECT ####")
    slope_aspect(dem_file, level_folder, example_fh)

    #### LULC ####
    print(f"\n#### LULC ####")
    if "GLOBCOVER" in source_selection["LULC"]:
        raw_lulc_file = c.Globcover.Landuse(*dl_args[:3])
        raw_lulc_files = [(year, raw_lulc_file) for year in range(sdate.year, edate.year + 1)]
    elif "WAPOR" in source_selection["LULC"]:
        raw_lulc_files = c.WAPOR.Get_Layer(*dl_args[:3], sdate.strftime("%Y-01-01"), edate.strftime("%Y-12-31"), 'L1_LCC_A')
        years = np.array([dat.strptime(os.path.split(fp)[-1], "L1_LCC_A_WAPOR_YEAR_%Y.%m.%d.tif").year for fp in raw_lulc_files])
        raw_lulc_files = [(year, raw_lulc_files[np.argmin(np.abs(years - year))]) for year in range(sdate.year, edate.year + 1)]        

    lulc_values = g.landcover_converter.get_lulc_values()

    lulc_file_template = unraw_filepaths(None, level_folder, "{var}_{year}", static = True)[0]
    for year, raw_file in raw_lulc_files:
        for key, replace_values in lulc_values[source_selection["LULC"][0]].items():
            # print(year, raw_file, key)
            unraw_replace_values(raw_file, lulc_file_template.format(var = key, year = year), replace_values, example_fh)

#%%
    #### METEO ####
    print(f"\n#### METEO ####")
    if "MERRA2" in source_selection["METEO"]:
        meteo_vars = ['t2m', 'u2m', 'v2m', 'q2m', 'tpw', 'ps', 'slp']
        raw_meteo_files = c.MERRA.daily_MERRA2(*dl_args, meteo_vars)
        raw_meteo_files = {**c.MERRA.daily_MERRA2(*dl_args, ['t2m'], data_type = ["min", "max"]), **raw_meteo_files}
        meteo_name_convertor = {'t2m-max': "t_air_max_24", 't2m-min': "t_air_min_24", 't2m': "t_air_24", 
                                'u2m': "u2m_24", 'v2m': "v2m_24",'q2m': "qv_24",'slp': "p_air_24_0"}
    elif "GEOS5" in source_selection["METEO"]:
        meteo_vars = ['t2m', 'u2m', 'v2m', 'qv2m', 'tqv', 'ps', 'slp']
        raw_meteo_files = c.GEOS.daily(*dl_args, meteo_vars)
        meteo_name_convertor = {'t2m': "t_air_24", 't2m-min': "t_air_min_24", 't2m-max': "t_air_max_24",  
                                'u2m': "u2m_24", 'v2m': "v2m_24",'slp': "p_air_24_0",'qv2m': "qv_24",}

    for var_name, raw_files in raw_meteo_files.items():

        print(f"## {var_name} ##")

        if "t2m" in var_name:
            raw_files = [g.lapse_rate.lapse_rate_temperature(x, dem_file) for x in raw_files if "_K_" in x]
        if var_name not in meteo_name_convertor.keys():
            continue
        units = {"t_air_max_24": "°C", "t_air_min_24":"°C", "t_air_24":"°C", 
          "u2m_24":"m/s", "v2m_24":"m/s","qv_24":"kg/kg","p_air_24_0":"kPa"}

        cmeta = {
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "linear",
            "var_name": meteo_name_convertor[var_name],
            "var_unit": units[meteo_name_convertor[var_name]],
        }
        ds = g.compositer.main(cmeta, [raw_files], epochs_info, temp_folder, example_ds, diagnostics = diagnostics)
        all_files[meteo_name_convertor[var_name]] = unraw_all(meteo_name_convertor[var_name], ds, unraw_file_templates, example_geoinfo)

#%%
    #### SE_ROOT ####
    ds_lst, ds_meteo, ds_ndvi, ds_temperature = pywapor.pre_se_root.main(project_folder, startdate, enddate, latlim, lonlim, level = source_selection)
    raw_se_root_files = pywapor.se_root.main(level_folder, ds_lst, ds_meteo, ds_ndvi, 
                                        ds_temperature, example_ds, example_geoinfo)

    cmeta = {
        "composite_type": 0.85,
        "temporal_interp": False,
        "spatial_interp": "nearest",
        "var_name": "se_root",
        "var_unit": "-",
    }
    ds = g.compositer.main(cmeta, [raw_se_root_files], epochs_info, temp_folder, example_ds, diagnostics = diagnostics)
    all_files["se_root"] = unraw_all(cmeta["var_name"], ds, unraw_file_templates, example_geoinfo)

#%%
    #### LAT LON ####
    print(f"\n#### LAT LON ####")
    lat_lon(example_ds, level_folder, example_geoinfo)

    #### SOLAR RADIATION ####
    print(f"\n#### SOLAR RADIATION ####")
    if "MERRA2" in source_selection["TRANS"]:
        raw_ra24_files = c.MERRA.daily_MERRA2(*dl_args, ['swgnet'])['swgnet']

    cmeta = {
        "composite_type": "mean",
        "temporal_interp": "linear",
        "spatial_interp": "linear",
        "var_name": "ra_24",
        "var_unit": "-",
    }
    ds = g.compositer.main(cmeta, [raw_ra24_files], epochs_info, temp_folder, example_ds, diagnostics = diagnostics)
    all_files["ra_24"] = unraw_all(cmeta["var_name"], ds, unraw_file_templates, example_geoinfo)

    #### TEMP. AMPLITUDE #### # TODO add source selection and remove year data
    print(f"\n#### TEMP. AMPLITUDE ####")
    raw_temp_ampl_file = os.path.join(raw_folder, "GLDAS", "Temp_Amplitudes_global.tif")
    download_file_from_google_drive("1pqZnCn-1xkUC7o1csG24hwg22fV57gCH", raw_temp_ampl_file)
    temp_ampl_file_template = unraw_filepaths(None, level_folder, "t_amp_year_{year}", static = True)[0]
    raw_temp_ampl_files = [(year, raw_temp_ampl_file) for year in range(sdate.year, edate.year + 1)]
    for year, raw_file in raw_temp_ampl_files:
        unraw(raw_file, temp_ampl_file_template.format(year = year), example_fh, 6)

    #### METADATA ####
    metadata = dict()
    metadata["pywapor_version"] = pywapor.__version__
    metadata["created"] = dat.now().strftime("%m/%d/%Y, %H:%M:%S")
    metadata["template_file"] = example_fh
    metadata["geotransform"] = "[{0}, {1}, {2}, {3}, {4}, {5}]".format(*PF.get_geoinfo(example_fh)[0])
    metadata["resolution"] = "[{0}, {1}]".format(*PF.get_geoinfo(example_fh)[2:])
    metadata["inputs"] = dict()
    metadata["inputs"]["latlim"] = "[{0}, {1}]".format(*latlim)
    metadata["inputs"]["lonlim"] = "[{0}, {1}]".format(*lonlim)
    metadata["inputs"]["startdate"] = startdate
    metadata["inputs"]["enddate"] = enddate
    metadata["inputs"]["source_selection_name"] = level
    metadata["inputs"]["project_folder"] = project_folder
    metadata["sources"] = source_selection
    json_file = os.path.join(level_folder, f"metadata_{level}.json")
    with open(json_file, 'w+') as f:
        json.dump(metadata, f, indent = 4 )

    for fh in glob.glob(os.path.join(temp_folder, "*.nc")):
        os.remove(fh)
    if len(os.listdir(temp_folder)) == 0:
        shutil.rmtree(temp_folder)

    os.chdir(project_folder)

    print("\n#################")
    print("## PRE_ET_LOOK ##")
    print("##### DONE ######\n")
#%%
    return all_files

def unraw_all(var_name, ds, unraw_file_templates, example_geoinfo):
    unrawed_files = list()
    assert ds.epoch.size == len(unraw_file_templates)
    for i, fh in enumerate(unraw_file_templates):
        if not -9999 in ds["epoch"].values:
            fh_date = pd.Timestamp(dat.strptime(fh.split("_")[-1], "%Y%m%d.tif"))
            assert fh_date == pd.Timestamp(ds.epoch_starts.isel(epoch = i).values)
        array = np.copy(ds.composite.isel(epoch = i).values)
        real_fh = fh.format(var = var_name)
        PF.Save_as_tiff(real_fh, array, example_geoinfo[0], example_geoinfo[1])
        unrawed_files.append(real_fh)
    ds.close()
    ds = None
    return unrawed_files

def select_template(fhs):
    fhs = [val for sublist in fhs for val in sublist]

    sizes = [gdal.Open(fh).RasterXSize * gdal.Open(fh).RasterYSize for fh in fhs]
    idx = np.argmax(sizes)

    example_fh = fhs[idx]
    example_ds = xr.open_dataset(example_fh).isel(band = 0).drop_vars(["band", "spatial_ref"]).rename({"x": "lon", "y": "lat"})
    example_geoinfo = PF.get_geoinfo(example_fh)

    return example_fh, example_ds, example_geoinfo

def slope_aspect(dem_file, project_folder, template_file):

    slope_file = unraw_filepaths(None, project_folder, "slope_deg", static = True)[0]
    aspect_file = unraw_filepaths(None, project_folder, "aspect_deg", static = True)[0]

    if not os.path.exists(slope_file) or not os.path.exists(aspect_file):

        dem = PF.open_as_array(dem_file)

        # constants
        geo_ex, proj_ex, size_x_ex, size_y_ex = PF.get_geoinfo(template_file)
        dlat, dlon = PF.calc_dlat_dlon(geo_ex, size_x_ex, size_y_ex)            

        pixel_spacing = (np.nanmean(dlon) + np.nanmean(dlat)) / 2
        rad2deg = 180.0 / np.pi  # Factor to transform from rad to degree

        # The slope output raster map contains slope values, stated in degrees of inclination from the horizontal
        # Calculate slope
        x, y = np.gradient(dem, pixel_spacing, pixel_spacing)
        hypotenuse_array = np.hypot(x,y)
        slope = np.arctan(hypotenuse_array) * rad2deg

        # calculate aspect
        aspect = np.arctan2(y/pixel_spacing, -x/pixel_spacing) * rad2deg
        aspect = 180 + aspect

        # Save as tiff files
        PF.Save_as_tiff(slope_file, slope, geo_ex, proj_ex)
        PF.Save_as_tiff(aspect_file, aspect, geo_ex, proj_ex)

    return slope_file, aspect_file

def lat_lon(ds, project_folder, example_geoinfo):

    lat_file = unraw_filepaths(None, project_folder, "lat_deg", static = True)[0]
    lon_file = unraw_filepaths(None, project_folder, "lon_deg", static = True)[0]

    lat_deg = np.rot90(np.tile(ds.lat, ds.lon.size).reshape(ds.band_data.shape[::-1]), 3)
    lon_deg = np.tile(ds.lon, ds.lat.size).reshape(ds.band_data.shape)
    if not os.path.exists(lon_file):
        PF.Save_as_tiff(lon_file, lon_deg, example_geoinfo[0], example_geoinfo[1])
    if not os.path.exists(lat_file):
        PF.Save_as_tiff(lat_file, lat_deg, example_geoinfo[0], example_geoinfo[1])

    return lat_file, lon_file

def unraw_filepaths(periods_start, project_folder, var, static = False):
    filepaths = list()
    if not static:
        for date in periods_start:
            date_str = pd.Timestamp(date).strftime("%Y%m%d")
            fp = os.path.join(project_folder,
                            date_str, f"{var}_{date_str}.tif")
            filepaths.append(fp)
    else:
        fp = os.path.join(project_folder,
                        "static", f"{var}.tif")
        filepaths.append(fp)   
    return filepaths

def create_dates(sdate, edate, period_length):
    if isinstance(sdate, str):
        sdate = dat.strptime(sdate, "%Y-%m-%d")
    if isinstance(edate, str):
        edate = dat.strptime(edate, "%Y-%m-%d")
    if period_length == "DEKAD":
        dekad1 = pd.date_range(sdate - pd.to_timedelta(sdate.day - 1, unit='d'), edate, freq = "MS")
        dekad2 = dekad1 + pd.Timedelta(10, unit='D')
        dekad3 = dekad1 + pd.Timedelta(20, unit='D')
        dates = np.sort(np.array([dekad1, dekad2, dekad3]).flatten())
        no_periods = dates.size - 1
    else:
        days = (edate - sdate).days
        no_periods = int(days // period_length + 1)
        dates = pd.date_range(sdate, periods = no_periods + 1 , freq = f"{period_length}D")
    periods_start = dates[:-1]
    mask = np.all([periods_start < np.datetime64(edate), 
                   periods_start >= np.datetime64(sdate)], axis = 0)
    epochs = np.arange(0, no_periods, 1)[mask]
    periods_end = dates[1:][mask]
    periods_start = periods_start[mask]
    return epochs, periods_start, periods_end

def unraw(raw_file, unraw_file, template_file, method):
    if not os.path.exists(unraw_file) and os.path.exists(raw_file):
        geo_ex, proj_ex = PF.get_geoinfo(template_file)[0:2]
        array = PF.reproj_file(raw_file, template_file, method)
        PF.Save_as_tiff(unraw_file, array, geo_ex, proj_ex)

def unraw_replace_values(raw_file, unraw_file, replace_values, template_file):
    if not os.path.exists(unraw_file) and os.path.exists(raw_file):
        array = PF.reproj_file(raw_file, template_file, 1)
        replaced_array = np.ones_like(array) * np.nan
        for key, value in replace_values.items():
            replaced_array[array == key] = value
        geo_ex, proj_ex = PF.get_geoinfo(template_file)[0:2]
        PF.Save_as_tiff(unraw_file, replaced_array, geo_ex, proj_ex)

def download_file_from_google_drive(id, destination):
    
    if os.path.isfile(destination):
        return

    folder = os.path.split(destination)[0]
    if not os.path.exists(folder):
        os.makedirs(folder)

    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id }, stream = True)
    token = get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    save_response_content(response, destination) 

    return           

def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value
    return None

def save_response_content(response, destination):
    CHUNK_SIZE = 32768
    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)

# if __name__ == "__main__":

    # project_folder = r"/Volumes/Data/FAO/WaPOR_vs_pyWaPOR/pyWAPOR_v1"
    # latlim = [28.9, 29.7]
    # lonlim = [30.2, 31.2]
    # startdate = "2021-07-01"
    # enddate = "2021-07-10"
    # composite_length = "DEKAD"

    # level = {
    #     "METEO": ["GEOS5"],
    #     "NDVI": ["MOD13", "MYD13", "PROBAV"],
    #     "ALBEDO": ["PROBAV"],
    #     "LST": ["MOD11", "MYD11"],
    #     "LULC": ["WAPOR"],
    #     "DEM": ["SRTM"],
    #     "PRECIPITATION": ["CHIRPS"],
    #     "TRANS": ["MERRA2"],
    # }
    # level = "level_2"

    # diagnostics = { # label          # lat      # lon
    #                 "water":	    (29.44977,	30.58215),
    #                 "desert":	    (29.12343,	30.51222),
    #                 "agriculture":	(29.32301,	30.77599),
    #                 "urban":	    (29.30962,	30.84109),
    #                 }
    # diagnostics = None

    # all_files = main(project_folder, startdate, enddate, latlim, lonlim, level = level, 
    #     diagnostics = diagnostics, composite_length = composite_length)


# -*- coding: utf-8 -*-
"""
WaterSat
author: Tim Martijn Hessels
Created on Tue Feb 26 08:17:10 2019
"""

def Globcover_LM(version = '1.0'):
    ETlook_LM = {
    11: 1,   #Post-flooding or irrigated croplands (or aquatic)
    14:	1,   #Rainfed croplands
    20:	1,   #Mosaic cropland (50-70%) / vegetation (grassland/shrubland/forest) (20-50%)
    30:	1,   #Mosaic vegetation (grassland/shrubland/forest) (50-70%) / cropland (20-50%) 
    40:	1,   #Closed to open (>15%) broadleaved evergreen or semi-deciduous forest (>5m)
    50:	1,   #Closed (>40%) broadleaved deciduous forest (>5m)
    60:	1,   #Open (15-40%) broadleaved deciduous forest/woodland (>5m)
    70:	1,   #Closed (>40%) needleleaved evergreen forest (>5m)
    90:	1,   #Open (15-40%) needleleaved deciduous or evergreen forest (>5m)
    100: 1,   	#Closed to open (>15%) mixed broadleaved and needleleaved forest (>5m)
    110: 1,   	#Mosaic forest or shrubland (50-70%) / grassland (20-50%)
    120: 1,   	#Mosaic grassland (50-70%) / forest or shrubland (20-50%) 
    130: 1,  	#Closed to open (>15%) (broadleaved or needleleaved, evergreen or deciduous) shrubland (<5m)
    140: 1,  	#Closed to open (>15%) herbaceous vegetation (grassland, savannas or lichens/mosses)
    150: 1,  	#Sparse (<15%) vegetation
    160: 1,  	#Closed to open (>15%) broadleaved forest regularly flooded (semi-permanently or temporarily) - Fresh or brackish water
    170: 1,  	#Closed (>40%) broadleaved forest or shrubland permanently flooded - Saline or brackish water
    180: 1,  	#Closed to open (>15%) grassland or woody vegetation on regularly flooded or waterlogged soil - Fresh, brackish or saline water
    190: 3,  	#Artificial surfaces and associated areas (Urban areas >50%)
    200: 1,  	#Bare areas
    210: 2,  	#Water bodies
    220: 1,  	#Permanent snow and ice
    230: 0  	#No data (burnt areas, clouds,…)
    }

    Classes_LM =dict()
    Classes_LM['1.0'] = ETlook_LM

    return Classes_LM[version]


def Globcover_MaxObs(version = '1.0'):
    ETlook_Classes = {
    11: 4.0,   #Post-flooding or irrigated croplands (or aquatic)
    14:	4.0,   #Rainfed croplands
    20:	2.0,   #Mosaic cropland (50-70%) / vegetation (grassland/shrubland/forest) (20-50%)
    30:	3.5,   #Mosaic vegetation (grassland/shrubland/forest) (50-70%) / cropland (20-50%) 
    40:	0.1,   #Closed to open (>15%) broadleaved evergreen or semi-deciduous forest (>5m)
    50:	0.6,   #Closed (>40%) broadleaved deciduous forest (>5m)
    60:	1.2,   #Open (15-40%) broadleaved deciduous forest/woodland (>5m)
    70:	2.0,   #Closed (>40%) needleleaved evergreen forest (>5m)
    90:	5.0,   #Open (15-40%) needleleaved deciduous or evergreen forest (>5m)
    100: 8.0,   	#Closed to open (>15%) mixed broadleaved and needleleaved forest (>5m)
    110: 2.0,   	#Mosaic forest or shrubland (50-70%) / grassland (20-50%)
    120: 8.0,   	#Mosaic grassland (50-70%) / forest or shrubland (20-50%) 
    130: 4.0,  	#Closed to open (>15%) (broadleaved or needleleaved, evergreen or deciduous) shrubland (<5m)
    140: 2.0,  	#Closed to open (>15%) herbaceous vegetation (grassland, savannas or lichens/mosses)
    150: 1.0,  	#Sparse (<15%) vegetation
    160: 0.3,  	#Closed to open (>15%) broadleaved forest regularly flooded (semi-permanently or temporarily) - Fresh or brackish water
    170: 6.0,  	#Closed (>40%) broadleaved forest or shrubland permanently flooded - Saline or brackish water
    180: 10,  	#Closed to open (>15%) grassland or woody vegetation on regularly flooded or waterlogged soil - Fresh, brackish or saline water
    190: 10,  	#Artificial surfaces and associated areas (Urban areas >50%)
    200: 0.1,  	#Bare areas
    210: 0.1,  	#Water bodies
    220: 0.1,  	#Permanent snow and ice
    230: 0  	#No data (burnt areas, clouds,…)
    }

    Classes_MaxObs =dict()
    Classes_MaxObs['1.0'] = ETlook_Classes

    return Classes_MaxObs[version]


def Globcover_Bulk(version = '1.0'):
    ETlook_Classes = {
    11: 200,   #Post-flooding or irrigated croplands (or aquatic)
    14:	200,   #Rainfed croplands
    20:	150,   #Mosaic cropland (50-70%) / vegetation (grassland/shrubland/forest) (20-50%)
    30:	150,   #Mosaic vegetation (grassland/shrubland/forest) (50-70%) / cropland (20-50%) 
    40:	100,   #Closed to open (>15%) broadleaved evergreen or semi-deciduous forest (>5m)
    50:	120,   #Closed (>40%) broadleaved deciduous forest (>5m)
    60:	100,   #Open (15-40%) broadleaved deciduous forest/woodland (>5m)
    70:	150,   #Closed (>40%) needleleaved evergreen forest (>5m)
    90:	180,   #Open (15-40%) needleleaved deciduous or evergreen forest (>5m)
    100: 175,   	#Closed to open (>15%) mixed broadleaved and needleleaved forest (>5m)
    110: 150,   	#Mosaic forest or shrubland (50-70%) / grassland (20-50%)
    120: 350,   	#Mosaic grassland (50-70%) / forest or shrubland (20-50%) 
    130: 175,  	#Closed to open (>15%) (broadleaved or needleleaved, evergreen or deciduous) shrubland (<5m)
    140: 250,  	#Closed to open (>15%) herbaceous vegetation (grassland, savannas or lichens/mosses)
    150: 150,  	#Sparse (<15%) vegetation
    160: 250,  	#Closed to open (>15%) broadleaved forest regularly flooded (semi-permanently or temporarily) - Fresh or brackish water
    170: 200,  	#Closed (>40%) broadleaved forest or shrubland permanently flooded - Saline or brackish water
    180: 300,  	#Closed to open (>15%) grassland or woody vegetation on regularly flooded or waterlogged soil - Fresh, brackish or saline water
    190: 100,  	#Artificial surfaces and associated areas (Urban areas >50%)
    200: 100,  	#Bare areas
    210: 100,  	#Water bodies
    220: 100,  	#Permanent snow and ice
    230: 0  	#No data (burnt areas, clouds,…)
    }

    Classes_Bulk =dict()
    Classes_Bulk['1.0'] = ETlook_Classes

    return Classes_Bulk[version]

def Globcover_LUEmax(version = '1.0'):
    ETlook_Classes = {
    11: 3.04,   #Post-flooding or irrigated croplands (or aquatic)
    14:	3.04,   #Rainfed croplands
    20:	3.04,   #Mosaic cropland (50-70%) / vegetation (grassland/shrubland/forest) (20-50%)
    30:	3.04,   #Mosaic vegetation (grassland/shrubland/forest) (50-70%) / cropland (20-50%) 
    40:	2.45,   #Closed to open (>15%) broadleaved evergreen or semi-deciduous forest (>5m)
    50:	2.66,   #Closed (>40%) broadleaved deciduous forest (>5m)
    60:	2.66,   #Open (15-40%) broadleaved deciduous forest/woodland (>5m)
    70:	2.5,   #Closed (>40%) needleleaved evergreen forest (>5m)
    90:	2.5,   #Open (15-40%) needleleaved deciduous or evergreen forest (>5m)
    100: 2.54,   	#Closed to open (>15%) mixed broadleaved and needleleaved forest (>5m)
    110: 2.54,   	#Mosaic forest or shrubland (50-70%) / grassland (20-50%)
    120: 2.54,   	#Mosaic grassland (50-70%) / forest or shrubland (20-50%) 
    130: 2.54,  	#Closed to open (>15%) (broadleaved or needleleaved, evergreen or deciduous) shrubland (<5m)
    140: 2.3,  	#Closed to open (>15%) herbaceous vegetation (grassland, savannas or lichens/mosses)
    150: 1.57,  	#Sparse (<15%) vegetation
    160: 2.66,  	#Closed to open (>15%) broadleaved forest regularly flooded (semi-permanently or temporarily) - Fresh or brackish water
    170: 1.57,  	#Closed (>40%) broadleaved forest or shrubland permanently flooded - Saline or brackish water
    180: 1.57,  	#Closed to open (>15%) grassland or woody vegetation on regularly flooded or waterlogged soil - Fresh, brackish or saline water
    190: 1.57,  	#Artificial surfaces and associated areas (Urban areas >50%)
    200: 1.57,  	#Bare areas
    210: 0,  	#Water bodies
    220: 0,  	#Permanent snow and ice
    230: 0  	#No data (burnt areas, clouds,…)
    }

    Classes_LUEmax =dict()
    Classes_LUEmax['1.0'] = ETlook_Classes

    return Classes_LUEmax[version]



def WAPOR_MaxObs(version = '1.0'):
    ETlook_Classes_v1 = {
    33: 8.0,   #Natural Vegetation
    41:	2.0,   #Cropland, Rainfed or Fallow
    42:	4.0,   #Cropland, Irrigated or Under Water Management
    50:	10,   #Artificial Surfaces
    60:	0.1,   #Bare Soil Or Sparse Vegetation
    80:	0.1   #Water
    }

    ETlook_Classes_v2 = {
    20.0:1.0,
    30.0:2.0,
    42.0:1.5,
    43.0:1.5,
    50.0:10.0,
    60.0:0.1,
    80.0:0.1,
    90.0:2.0,
    116.0:5.0,
    126.0:3.0,
    33: 8.0,   #Natural Vegetation
    41:	2.0,   #Cropland, Rainfed or Fallow
    }

    Classes_MaxObs =dict()
    Classes_MaxObs['1.0'] = ETlook_Classes_v1
    Classes_MaxObs['2.0'] = ETlook_Classes_v2

    return Classes_MaxObs[version]

def WAPOR_Bulk(version = '1.0'):
    ETlook_Classes_v1 = {
    33: 175,   #Natural Vegetation
    41:	150,   #Cropland, Rainfed or Fallow
    42:	200,   #Cropland, Irrigated or Under Water Management
    50:	400,   #Artificial Surfaces
    60:	100,   #Bare Soil Or Sparse Vegetation
    80:	100   #Water
    }

    ETlook_Classes_v2 = {
    20.0:175,
    30.0:175,
    42.0:125,
    43.0:125,
    50.0:400,
    60.0:100,
    80.0:100,
    90.0:150,
    116.0:180,
    126.0:250,
    33: 175,   #Natural Vegetation
    41:	150,   #Cropland, Rainfed or Fallow
    }

    Classes_Bulk =dict()
    Classes_Bulk['1.0'] = ETlook_Classes_v1
    Classes_Bulk['2.0'] = ETlook_Classes_v2

    return Classes_Bulk[version]


def WAPOR_LM(version = '1.0'):
    ETlook_Classes_v1 = {
    33: 1,   #Natural Vegetation
    41:	1,   #Cropland, Rainfed or Fallow
    42:	1,   #Cropland, Irrigated or Under Water Management
    50:	3,   #Artificial Surfaces
    60:	1,   #Bare Soil Or Sparse Vegetation
    80:	2   #Water
    }

    ETlook_Classes_v2 = {
    20.0:1,
    30.0:1,
    42.0:1,
    43.0:1,
    50.0:3,
    60.0:1,
    80.0:2,
    90.0:1,
    116.0:1,
    126.0:1,
    33: 1,   #Natural Vegetation
    41:	1,   #Cropland, Rainfed or Fallow
    }

    Classes_Bulk =dict()
    Classes_Bulk['1.0'] = ETlook_Classes_v1
    Classes_Bulk['2.0'] = ETlook_Classes_v2

    return Classes_Bulk[version]


def WAPOR_LUEmax(version = '1.0'):
    ETlook_Classes = {
    33: 2.66,   #Natural Vegetation
    41:	3.04,   #Cropland, Rainfed or Fallow
    42:	3.04,   #Cropland, Irrigated or Under Water Management
    50:	1.57,   #Artificial Surfaces
    60:	2.66,   #Bare Soil Or Sparse Vegetation
    80:	0   #Water
    }

    Classes_LUEmax =dict()
    Classes_LUEmax['1.0'] = ETlook_Classes

    return Classes_LUEmax[version]

def get_lulc_values():
    lulc_values = {"GLOBCOVER": {"land_mask": Globcover_LM(),
                                "rs_min": Globcover_Bulk(),
                                "z_obst_max": Globcover_MaxObs(),
                                "lue_max": Globcover_LUEmax()},
                    "WAPOR":     {"land_mask": WAPOR_LM(version = "2.0"),
                                "rs_min": WAPOR_Bulk(version = "2.0"),
                                "z_obst_max": WAPOR_MaxObs(version = "2.0"),
                                "lue_max": WAPOR_LUEmax(version = "1.0")}
                            }

    return lulc_values


# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/CHIRPS
"""

# import general python modules
import os
from osgeo import gdal
import numpy as np
import pandas as pd
from ftplib import FTP
import datetime

def DownloadData(Dir, Startdate, Enddate, latlim, lonlim, Waitbar):
    """
    This function downloads CHIRPS daily or monthly data

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax] (values must be between -50 and 50)
    lonlim -- [xmin, xmax] (values must be between -180 and 180)
    Waitbar -- 1 (Default) will print a waitbar
    cores -- The number of cores used to run the routine. It can be 'False'
             to avoid using parallel computing routines.
    TimeCase -- String equal to 'daily' or 'monthly'
    """
	
    # Define timestep for the timedates
    output_folder = os.path.join(Dir, "CHIRPS", "Precipitation")

    # make directory if it not exists
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

	# check time variables
    if not Startdate:
        Startdate = pd.Timestamp('1981-01-01')
    if not Enddate:
        Enddate = pd.Timestamp('Now')

    # Create days
    Dates = pd.date_range(Startdate, Enddate, freq='D')

    # Create Waitbar
    if Waitbar == 1:
        import pywapor.general.waitbar_console as WaitbarConsole
        total_amount = len(Dates)
        amount = 0
        WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)

    # Check space variables
    if latlim[0] < -50 or latlim[1] > 50:
        print('Latitude above 50N or below 50S is not possible.'
               ' Value set to maximum')
        latlim[0] = np.max(latlim[0], -50)
        latlim[1] = np.min(lonlim[1], 50)
    if lonlim[0] < -180 or lonlim[1] > 180:
        print('Longitude must be between 180E and 180W.'
               ' Now value is set to maximum')
        lonlim[0] = np.max(latlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)

    # Define IDs
    yID = 2000 - np.int16(np.array([np.ceil((latlim[1] + 50)*20),
                                    np.floor((latlim[0] + 50)*20)]))
    xID = np.int16(np.array([np.floor((lonlim[0] + 180)*20),
                             np.ceil((lonlim[1] + 180)*20)]))

    # Pass variables to parallel function and run
    args = [output_folder, xID, yID, lonlim, latlim]
    for Date in Dates:
        RetrieveData(Date, args)
        
        if Waitbar == 1:
            amount += 1
            WaitbarConsole.printWaitBar(amount, total_amount, prefix = 'Progress:', suffix = 'Complete', length = 50)
    results = True
    
    # remove raw files
    import glob
    os.chdir(output_folder)
    try:
        files_raw = glob.glob("chirps-v2.0*.tif")
        for file_raw in files_raw:
            outfilename = os.path.join(output_folder, file_raw)
            # delete old tif file
            os.remove(outfilename)
    except:
        pass        
        
    return results

def RetrieveData(Date, args):
    """
    This function retrieves CHIRPS data for a given date from the
    ftp://chg-ftpout.geog.ucsb.edu server.

    Keyword arguments:
    Date -- 'yyyy-mm-dd'
    args -- A list of parameters defined in the DownloadData function.
    """
    
    # WA+ modules
    import pywapor.general.processing_functions as PF  

    # Argument
    [output_folder, xID, yID, lonlim, latlim] = args

    # Define output
    DirFileEnd = os.path.join(output_folder,'P_CHIRPS.v2.0_mm-day-1_daily_%s.%02s.%02s.tif' %(Date.strftime('%Y'), Date.strftime('%m'), Date.strftime('%d')))

    if not os.path.exists(DirFileEnd):

        # open ftp server
        ftp = FTP("ftp.chc.ucsb.edu", "", "")
        ftp.login()
    
    	# Define FTP path to directory
        pathFTP = 'pub/org/chg/products/CHIRPS-2.0/global_daily/tifs/p05/%s/' %Date.strftime('%Y')
    
        # find the document name in this directory
        ftp.cwd(pathFTP)
        listing = []
    
    	# read all the file names in the directory
        ftp.retrlines("LIST", listing.append)
    
    	# create all the input name (filename) and output (outfilename, filetif, DiFileEnd) names
        filename = 'chirps-v2.0.%s.%02s.%02s.tif.gz' %(Date.strftime('%Y'), Date.strftime('%m'), Date.strftime('%d'))
        outfilename = os.path.join(output_folder,'chirps-v2.0.%s.%02s.%02s.tif' %(Date.strftime('%Y'), Date.strftime('%m'), Date.strftime('%d')))

        # download the global rainfall file

        # unzip the file if necessary
        # Files after 2021-6-1 are not zipped for some reason (Bert, 29-9-2021)
        if Date >= datetime.datetime(2021, 6, 1):
            filename = filename.replace(".gz", "")
            with open(outfilename, "wb") as lf:
                ftp.retrbinary("RETR " + filename, lf.write, 8192)
        else:
            local_filename = os.path.join(output_folder, filename)
            with open(local_filename, "wb") as lf:
                ftp.retrbinary("RETR " + filename, lf.write, 8192) 
            PF.Extract_Data_gz(local_filename, outfilename)

        # open tiff file
        dest = gdal.Open(outfilename)
        dataset = dest.GetRasterBand(1).ReadAsArray()

        # clip dataset to the given extent
        data = dataset[yID[0]:yID[1], xID[0]:xID[1]]
        data[data < 0] = -9999

        # save dataset as geotiff file
        geo = [lonlim[0], 0.05, 0, latlim[1], 0, -0.05]
        
        PF.Save_as_tiff(name=DirFileEnd, data=data, geo=geo, projection="WGS84")
    
    return True



import xarray as xr
import os
import numpy as np
import pandas as pd
import pywapor
import pywapor.general.processing_functions as PF
import pywapor.et_look_dev as ETLook_dev
import pywapor.et_look_v2 as ETLook_v2
import tqdm

def main(level_folder, ds_lst, ds_meteo, ds_ndvi, 
                    ds_temperature, example_ds, example_geoinfo, et_look_version = "v2"):

    print("\n#### SE_ROOT ####")

    # Version
    if et_look_version == "v2":
        ETLook = ETLook_v2
    elif et_look_version == "dev":
        ETLook = ETLook_dev

    # spatial interpolation
    ds_lst2 = ds_lst.interp_like(example_ds, method = "linear", kwargs={"fill_value": "extrapolate"},)
    ds_meteo2 = ds_meteo.interp_like(example_ds, method = "linear", kwargs={"fill_value": "extrapolate"},)
    ds_ndvi2 = ds_ndvi.interp_like(example_ds, method = "linear", kwargs={"fill_value": "extrapolate"},)
    ds_temperature2 = ds_temperature.interp_like(example_ds, method = "linear", kwargs={"fill_value": "extrapolate"},)

    # temporal interpolation
    ds_meteo3 = ds_meteo2.interp(time = ds_lst2.time, method = "nearest", kwargs={"fill_value": "extrapolate"},) # TODO download more meteo data and switch to linear
    ds_ndvi3 = ds_ndvi2.interp(time = ds_lst.time, method = "linear", kwargs={"fill_value": "extrapolate"},)
    ds_temperature3 = ds_temperature2.interp(time = ds_lst2.time, method = "nearest", kwargs={"fill_value": "extrapolate"},)

    req_vars = ['time', 'angle', 'lst', 'lon', 'lat', 'Pair_inst_0', 'Pair_inst',
                'qv_inst', 'tair_inst', 'wv_inst', 'v2m_inst', 'u2m_inst', 'ndvi']

    renames = {"Pair_inst_0": "p_air_0_i", "Pair_inst": "p_air_i", # TODO Make this renaming unnesecary
                "tair_inst": "t_air_i", "qv_inst": "qv_i", 
                "lon": "lon_deg", "wv_inst": "wv_i", "lat": "lat_deg"}

    ds_se_root = xr.merge([ds_lst2, ds_meteo3, ds_ndvi3, ds_temperature3])
    ds_se_root = ds_se_root.drop_vars([x for x in list(ds_se_root.variables) if x not in req_vars])
    ds_se_root = ds_se_root.rename({k: v for k, v in renames.items() if k in list(ds_se_root.variables)})
    ds_se_root["u_i"] = np.sqrt(ds_se_root["v2m_inst"]**2 + ds_se_root["u2m_inst"]**2)

    se_root_folder = os.path.join(level_folder, "SMC")
    if not os.path.exists(se_root_folder):
        os.mkdir(se_root_folder)

    files = list()

    print("--> Calculating se_root.")
    for t in tqdm.tqdm(ds_se_root.time.values): # TODO adjust pywapor.et_look.se_root to remove this forloop

        id = ds_se_root.sel(time = t)
        date = pd.Timestamp(t)

        dt_str = date.strftime("%Y.%m.%d.%H.%M")
        fn = f"SMC_-_-_inst_{dt_str}.tif"
        fh = os.path.join(se_root_folder, fn)

        if not os.path.exists(fh): # TODO also check if fh has correct geoinfo
            se_root_i = pywapor.et_look.se_root(id, None, ETLook, date, verbose = True)[0]["se_root"]
            PF.Save_as_tiff(fh, se_root_i.values, example_geoinfo[0], example_geoinfo[1])

        files.append(fh)

    return files


# -*- coding: utf-8 -*-
"""
Savitzky-Gloay NDVI filtering
author: Laust Færch @ DHI GRAS
Created on 2020/10/27

Based on the article:
Chen, J., Jönsson, P., Tamura, M., Gu, Z., Matsushita, B., & Eklundh, L. 2004. A simple method for reconstructing a
high-quality NDVI time-series data set based on the Savitzky-Golay filter.

"""


import numpy as np

from numba import jit
from scipy import interpolate
from scipy.signal import savgol_filter

def _interpolate_1d(data):
    # we need at least 2 non-nan elements
    if np.sum(~np.isnan(data) * 1) < 2:
        return data

    good = ~np.isnan(data)

    # scipy interpolation
    finterp = interpolate.interp1d(np.flatnonzero(good),
                                   data[good], kind='linear',
                                   fill_value=np.nan,
                                   bounds_error=False)

    yinterp = finterp(np.arange(data.shape[0]))

    return yinterp


@jit
def _prepare_smoothing(N0):
    # long term change trend fitting:
    m = list(np.array([4, 5, 6, 7]) * 2 + 1)
    d = [2, 3, 4]

    parameters = np.array(np.meshgrid(m, d)).T.reshape(-1, 2)  # not Numba supported

    long_term_trend = np.zeros((N0.shape[0], parameters.shape[0]))

    for i, param in enumerate(parameters):
        long_term_trend[:, i] = savgol_filter(N0, param[0], param[1])

    # find minimum square error
    square_error = (long_term_trend - N0[:, None]) ** 2
    min_square_error_arg = np.argmin(np.sum(square_error, axis=0))

    Ntr = long_term_trend[:, min_square_error_arg]

    # Determination of NDVI weights for each point (based on difference from original data)
    W = np.ones_like(N0)
    di = np.abs(N0 - Ntr)
    W_full = 1 - di / np.max(di)
    W[N0 < Ntr] = W_full[N0 < Ntr]

    return Ntr, W


@jit
def _recursive_smoothing(N0, Ntr, W, niter=0):
    max_iter = 5
    ndvi_stopping_difference = 0.05

    # generate new NDVI timeseries
    N1 = N0.copy()
    N1[N0 < Ntr] = Ntr[N0 < Ntr]

    # Fitting new trend using savgol (m=4, d=6)
    Nk = savgol_filter(N1, 9, 6)

    # calculate fitting-effect index (changed to max ndvi diff)
    max_diff = np.max(np.abs(Nk - N0))

    if (max_diff >= ndvi_stopping_difference) & (niter <= max_iter):
        N1 = _recursive_smoothing(Nk, Ntr, W, niter + 1)

    return N1


@jit
def _savgol_reconstruct_1d(N0):

    # dont smooth if all nan
    if np.all(np.isnan(N0)):
        return N0

    Ntr, W = _prepare_smoothing(N0)
    Nk = _recursive_smoothing(N0, Ntr, W)
    return Nk


def savgol_reconstruct(data, axis=0, invert=False):
    """
    :param data: np.array((y,x,t))
    NDVI/Albedo raster timeseries. Each image in timeseries is shape y,x.
    :param axis: int
    Axis for the time-dimension in the array. Filtering/interpolation will be apply along this axis
    :param invert: bool
    Inversion of the data and output. This is useful for albedo where bad cloud masking will force the values up,
    instead of down for NDVI.

    :return: y_smoothed: np.array(y,x,t)
    """
    if invert:
        data = data * -1

    print('interpolating...')
    data_interp = np.apply_along_axis(_interpolate_1d, axis, data)

    print('reconstructing...')
    data_recons = np.apply_along_axis(_savgol_reconstruct_1d, axis, data_interp)

    if invert:
        data_recons = data_recons * -1
        data_interp = data_interp * -1

    return data_recons, data_interp



import sys
from pywapor.collect.MYD13.DataAccess import DownloadData
from datetime import date
import glob
import os
import pywapor 

def main(Dir, latlim, lonlim, Startdate, Enddate, Waitbar = 1, 
        hdf_library = None, remove_hdf = 1, buffer_dates = False):
    """
    This function downloads MYD13 16-daily data for the specified time
    interval, and spatial extent.

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Startdate -- 'yyyy-mm-dd' or datetime.date
    Enddate -- 'yyyy-mm-dd' or datetime.date
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    username -- "" string giving the username of your NASA account (https://urs.earthdata.nasa.gov/)
    password -- "" string giving the password of your NASA account    
    Waitbar -- 1 (Default) will print a waitbar
    hdf_library -- string, if all the hdf files are already stored on computer
                    define directory to the data here
    remove_hdf -- 1 (Default), if 1 remove all the downloaded hdf files in the end    
    """
    if isinstance(Startdate, date):
        Startdate = Startdate.strftime("%Y-%m-%d")
    
    if isinstance(Enddate, date):
        Enddate = Enddate.strftime("%Y-%m-%d")

    username, password = pywapor.collect.get_pw_un.get("NASA")

    print('\nDownload 16-daily MYD13 NDVI data for period %s till %s' %(Startdate, Enddate))
    DownloadData(Dir, Startdate, Enddate, latlim, lonlim, username, 
                password, Waitbar, hdf_library, remove_hdf, buffer_dates = buffer_dates)

    return glob.glob(os.path.join(Dir, "MODIS", "MYD13", "*.tif"))


if __name__ == '__main__':
    main(sys.argv)


# -*- coding: utf-8 -*-
"""
Created on Wed Sep  9 08:28:08 2020

@author: timhe
"""
import os
import json
from cryptography.fernet import Fernet
import pywapor
import getpass
import sys

def create_key():

    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    filename = "secret.txt"
    filehandle = os.path.join(folder, filename)

    if os.path.exists(filehandle):
        os.remove(filehandle)
    
    f = open(filehandle,"w+")
    f.write(str(Fernet.generate_key().decode("utf-8")))
    f.close()

    return

def setup_account(account):

    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    filename = "secret.txt"
    filehandle = os.path.join(folder, filename)
    json_filehandle = os.path.join(folder, "keys.json")

    if not os.path.exists(filehandle):
        create_key()
    
    f = open(filehandle,"r")
    key = f.read()
    f.close()
    
    cipher_suite = Fernet(key.encode('utf-8'))

    if os.path.exists(json_filehandle):
        # open json file  
        with open(json_filehandle) as f:
            datastore = f.read()
        obj = json.loads(datastore)      
        f.close()
    else:
        obj = {}

    if account == "WAPOR":
        API_Key = input(f"{account} API token: ")
        API_Key_crypt = cipher_suite.encrypt(("%s" %API_Key).encode('utf-8'))
        obj[account] = [str(API_Key_crypt.decode("utf-8"))]
    else:
        account_name = input(f"{account} username: ")
        pwd = getpass.getpass(f"{account} password: ")            
        account_name_crypt = cipher_suite.encrypt(("%s" %account_name).encode('utf-8'))
        pwd_crypt = cipher_suite.encrypt(("%s" %pwd).encode('utf-8'))
        obj[account] = ([str(account_name_crypt.decode("utf-8")), str(pwd_crypt.decode("utf-8"))])

    # save extent in task
    with open(json_filehandle, 'w') as outfile:
        json.dump(obj, outfile)        

    if account == "NASA":
        succes = pywapor.general.tests.nasa_account()
        if succes:
            print("NASA account: working")
        else:
            to_be_removed = obj.pop("NASA")
            with open(json_filehandle, 'w') as outfile:
                json.dump(obj, outfile)
            sys.exit(f"Please fix your NASA account.")            
   
    return
    



import xarray as xr
import numpy as np
from datetime import datetime as dat
import pywapor.collect as c
import pywapor.general as g
import os
import pandas as pd
from datetime import time as datt
import tqdm
import warnings

def main(project_folder, startdate, enddate, latlim, lonlim, level = "level_1"):

    print("\n#################")
    print("## PRE_SE_ROOT ##")
    print("#################")

    # Disable Warnings
    warnings.filterwarnings('ignore')

    sdate = dat.strptime(startdate, "%Y-%m-%d").date()
    edate = dat.strptime(enddate, "%Y-%m-%d").date()

    #### Check if selected sources and dates are valid
    levels = g.variables.get_source_level_selections()

    if isinstance(level, dict):
        level_name = "custom"
        if "name" in level.keys():
            level_name = level.pop("name")
        levels[level_name] = level
        level = level_name

    source_selection = levels[level]
    succes = g.tests.check_source_selection(source_selection, sdate, edate)[1]
    assert succes, "invalid source_selection"

    raw_folder = os.path.join(project_folder, "RAW")
    temp_folder = os.path.join(project_folder, "temporary")

    dl_args = (raw_folder, latlim, lonlim, startdate, enddate)

    #### NDVI ####
    print(f"\n#### NDVI ####")
    raw_ndvi_files = list()
    # Order is important! PROBV gets priority over MOD13, and MOD13 over MYD13.
    if "PROBAV" in source_selection["NDVI"]:
        raw_ndvi_files.append(c.PROBAV.PROBAV_S5(*dl_args)[0])
    if "MOD13" in source_selection["NDVI"]:
        raw_ndvi_files.append(c.MOD13.NDVI(*dl_args, remove_hdf = 0))
    if "MYD13" in source_selection["NDVI"]:
        raw_ndvi_files.append(c.MYD13.NDVI(*dl_args))

    cmeta = {
        "composite_type": False,
        "temporal_interp": False,
        "spatial_interp": "nearest",
        "var_name": "NDVI",
        "var_unit": "-",
    }
    ds = g.compositer.main(cmeta, raw_ndvi_files, None, temp_folder, None, lean_output = False)
    ds_ndvi = ds.rename({"band_data": "ndvi"})

    #### LST ####
    print(f"\n#### LST ####")
    raw_lst_files = list()
    if "MOD11" in source_selection["LST"]:
        raw_lst_files.append(c.MOD11.LST(*dl_args))
    if "MYD11" in source_selection["LST"]:
        raw_lst_files.append(c.MYD11.LST(*dl_args))
    ds_lst = combine_lst(raw_lst_files)

    #### DEM ####
    print(f"\n#### DEM ####")
    if "SRTM" in source_selection["DEM"]:
        raw_dem_file = c.SRTM.DEM(*dl_args[:3])

    #### METEO ####
    print(f"\n#### METEO ####")
    if "MERRA2" in source_selection["METEO"]:
        freq = "H"
        periods, period_times = calc_periods(ds_lst.time.values, freq)
        ds_lst["periods"] = xr.DataArray([x[2] for x in periods], {"time": ds_lst.time})
        meteo_vars = ['t2m', 'u2m', 'v2m', 'q2m', 'tpw', 'ps', 'slp']
        waitbar = tqdm.tqdm(total = len(periods) * len(meteo_vars), delay = 10, initial = 1)
        for sd, ed, period in periods:
            c.MERRA.hourly_MERRA2(*dl_args[:3], sd, ed, meteo_vars, [int(period)])
    elif "GEOS5" in source_selection["METEO"]:
        freq = "3H"
        periods, period_times = calc_periods(ds_lst.time.values, freq)
        ds_lst["periods"] = xr.DataArray([x[2] for x in periods], {"time": ds_lst.time})
        meteo_vars = ['t2m', 'u2m', 'v2m', 'qv2m', 'tqv', 'ps', 'slp']
        waitbar = tqdm.tqdm(total = len(periods) * len(meteo_vars), delay = 10, initial = 1)
        for sd, ed, period in periods:
            c.GEOS.three_hourly(*dl_args[:3], sd, ed, meteo_vars, [int(period)], Waitbar = waitbar)

    #### INST. METEO ####
    raw_inst_meteo_paths = g.variables.get_raw_meteo_paths("inst")
    ds_temperature = None
    ds_meteo = None
    for var_name, folder in raw_inst_meteo_paths[source_selection["METEO"][0]].items():
        print(f"## {var_name} ##")
        for date, period in zip(ds_lst.time.values, ds_lst.periods.values):
            hour = int((period - 1) * {"3H": 3, "H": 1}[freq])
            hour_str = str(hour).zfill(2)
            date_str = pd.Timestamp(date).strftime("%Y.%m.%d")
            raw_file = os.path.join(*folder).format(raw_folder = raw_folder, date = date_str, hour = hour_str)
            if var_name == "tair_inst":
                raw_file = g.lapse_rate.lapse_rate_temperature(raw_file, raw_dem_file)
                meteo_datetime = dat.combine(pd.Timestamp(date).date(), period_times[period])
                da_var = xr.open_dataset(raw_file).squeeze("band").rename({"band_data": var_name, "x": "lon", "y": "lat"})[var_name]
                da_var = da_var.assign_coords({"time": meteo_datetime}).expand_dims("time", axis = 0)
                ds_temp = xr.Dataset({var_name: da_var})
                if isinstance(ds_temperature, type(None)):
                    ds_temperature = ds_temp
                elif np.datetime64(meteo_datetime) in ds_temperature.time:
                    ds_temperature = xr.merge([ds_temperature, ds_temp])
                else:
                    ds_temperature = xr.concat([ds_temperature, ds_temp], dim = "time")
            else:
                meteo_datetime = dat.combine(pd.Timestamp(date).date(), period_times[period])
                da_var = xr.open_dataset(raw_file).squeeze("band").rename({"band_data": var_name, "x": "lon", "y": "lat"})[var_name]
                da_var = da_var.assign_coords({"time": meteo_datetime}).expand_dims("time", axis = 0)
                ds_temp = xr.Dataset({var_name: da_var})
                if isinstance(ds_meteo, type(None)):
                    ds_meteo = ds_temp
                elif np.datetime64(meteo_datetime) in ds_meteo.time:
                    ds_meteo = xr.merge([ds_meteo, ds_temp])
                else:
                    ds_meteo = xr.concat([ds_meteo, ds_temp], dim = "time")

    print("\n#################")
    print("## PRE_SE_ROOT ##")
    print("##### DONE ######\n")

    return ds_lst, ds_meteo, ds_ndvi, ds_temperature

def calc_periods(times, freq):
    periods = list()

    for now_time in times:

        date = pd.Timestamp(now_time).date()
        starttime = dat(date.year, date.month, date.day, 0, 0)
        endtime = dat(date.year, date.month, date.day, 23, 59)

        offsets = {"3H": 90, "H": 30}

        DateTime = pd.date_range(starttime, endtime, freq=freq) + pd.offsets.Minute(offsets[freq])
        Time_nearest = min(DateTime, key=lambda DateTime: abs(DateTime - now_time))
        period = np.argwhere(DateTime ==Time_nearest)[0][0] + 1

        periods.append((starttime, endtime, period))

    period_times = {i+1: datt(time.hour, time.minute) for i, time in enumerate(DateTime)}

    return periods, period_times

def combine_lst(lst_files):
    ds = None

    for lst_db in lst_files:
        for date, files in lst_db.items():

            ds_time = xr.open_dataset(files[1]).squeeze("band").rename({"band_data": "time", "x": "lon", "y": "lat"})
            ds_angle = xr.open_dataset(files[2]).squeeze("band").rename({"band_data": "angle", "x": "lon", "y": "lat"})
            ds_lst = xr.open_dataset(files[0]).squeeze("band").rename({"band_data": "lst", "x": "lon", "y": "lat"})

            unique_times = np.unique(ds_time.time)
            unique_times = unique_times[np.isfinite(unique_times)]

            offset_GTM = int(round(np.median(ds_time.lon) * 24 / 360)) # TODO Make this spatial for areas spanning multiple timezones

            for time in unique_times:

                hours = np.floor(time).astype(int) - offset_GTM
                minutes = np.round((time - offset_GTM - hours) * 60).astype(int)
                dt = dat.combine(date, datt(hours, minutes))

                da_angle = ds_angle.angle.where(ds_time.time == time)
                da_angle = da_angle.assign_coords({"time": dt}).expand_dims("time", axis = 0)

                da_lst = ds_lst.lst.where(ds_time.time == time, np.nan)
                da_lst = da_lst.assign_coords({"time": dt}).expand_dims("time", axis = 0)

                if isinstance(ds, type(None)):
                    ds = xr.Dataset({"angle": da_angle, "lst": da_lst})
                else:
                    ds_temp = xr.Dataset({"angle": da_angle, "lst": da_lst})
                    ds = xr.concat([ds, ds_temp], dim = "time")

    ds.attrs["time"] = "<UTC> time"

    return ds


# -*- coding: utf-8 -*-
"""
Authors: Laust Færch
Module: Collect/PROBAV

Description:
This module preprocesses Landsat(7+8) data downloaded externaly

"""
from .PrepareLandsat import main as PrepareLandsat

__all__ = ["PrepareLandsat"]

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Created on Mon Jan  4 12:22:19 2021

@author: timhe
"""

import os
import sys
import shutil
import datetime
from osgeo import gdal
import requests
import glob
import pandas as pd
import numpy as np

import watertools
import watertools.General.data_conversions as DC
import watertools.General.raster_conversions as RC

import pywapor


def main(output_folder, Startdate, Enddate, latlim, lonlim, LS_WAPOR_folder_in, LS_WAPOR_folder_out, LandCover = "GlobCover", Short_Downwards_Radiation = "MSGCCP", RAW_folder = None):

    ############################ Get General inputs ###############################
    
    # Define the input folders
    folders_input_RAW = os.path.join(output_folder, "RAW_composite")
    folder_input_ETLook = os.path.join(output_folder, "ETLook_input_composite_LS")
  
    if RAW_folder != None:
        folders_input_RAW = RAW_folder       
  
    # Create folders if not exists
    if not os.path.exists(folders_input_RAW):
        os.makedirs(folders_input_RAW)
    if not os.path.exists(folder_input_ETLook):
        os.makedirs(folder_input_ETLook)
       
    # Define the dates    #!!! is now 8 daily, needs to be changed into dekadal
    Dates = watertools.Collect.MOD11.DataAccess.Make_TimeStamps(Startdate, Enddate)
    Timestep_composite = 8
        
    ########################### Download CHIRPS data ############################## 

    ######################## Download Rainfall Data ###############################
    # Download CHIRPS data
    watertools.Collect.CHIRPS.daily(folders_input_RAW, Startdate, Enddate, latlim, lonlim)

    ########################### Download DEM data #################################
    
    # Download DEM data
    watertools.Collect.DEM.SRTM(folders_input_RAW, latlim, lonlim)
    
    ############################ Download Landuse #################################
    if LandCover == "GlobCover":
        # Download Globcover data
        watertools.Collect.Globcover.Landuse(folders_input_RAW, latlim, lonlim)
        
    if LandCover == "WAPOR":
       
        # Download Globcover data
        watertools.Collect.WAPOR.Get_Layer(folders_input_RAW, "%s-01-01"%(Startdate.split("-")[0]), "%s-12-31"%(Enddate.split("-")[0]), latlim, lonlim, "L1_LCC_A")        

    ############################ Download Static Amplitude map #################################
    
    print("download temperature amplitudes from google drive")       
    output_folder_Tamp = os.path.join(folders_input_RAW, "GLDAS")
    if not os.path.exists(output_folder_Tamp):
        os.makedirs(output_folder_Tamp)
    
    T_amplitude_global_temp_filename = os.path.join(output_folder_Tamp, "Temp_Amplitudes_global.tif")
    if not os.path.exists(T_amplitude_global_temp_filename):
        pywapor.pre_et_look.download_file_from_google_drive("1pqZnCn-1xkUC7o1csG24hwg22fV57gCH", T_amplitude_global_temp_filename)

    ############### Loop over days for the dynamic data ###############################
    
    # Create the inputs of MODIS
    for Date in Dates:
        
        try:
            # Define output folder
            folder_input_ETLook_Date = os.path.join(folder_input_ETLook, "%d%02d%02d" %(Date.year, Date.month, Date.day))
            if not os.path.exists(folder_input_ETLook_Date):
                os.makedirs(folder_input_ETLook_Date)

            folder_input_ETLook_Static = os.path.join(folder_input_ETLook, "Static")
            if not os.path.exists(folder_input_ETLook_Static):
                os.makedirs(folder_input_ETLook_Static)


            ##########################################################################
            # Create Albedo, NDVI and SE root
            ##########################################################################    
            
            NDVI_file = os.path.join(folder_input_ETLook_Date, "NDVI_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))            
            if not os.path.exists(NDVI_file):
                
                input_format_NDVI = os.path.join(LS_WAPOR_folder_in, "{yyyy}{mm:02d}{dd:02d}", "NDVI_{yyyy}{mm:02d}{dd:02d}.tif")
                NDVI_comp, geo, proj = Calc_Composite_Indicators_LS(input_format_NDVI, Date, Timestep_composite, method = "mean")
                DC.Save_as_tiff(NDVI_file, NDVI_comp, geo, proj)
            
            ALBEDO_file = os.path.join(folder_input_ETLook_Date, "ALBEDO_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))
            if not os.path.exists(ALBEDO_file):
                
                input_format_ALBEDO = os.path.join(LS_WAPOR_folder_in, "{yyyy}{mm:02d}{dd:02d}", "ALBEDO_{yyyy}{mm:02d}{dd:02d}.tif")
                ALBEDO_comp, geo, proj = Calc_Composite_Indicators_LS(input_format_ALBEDO, Date, Timestep_composite, method = "mean")
                DC.Save_as_tiff(ALBEDO_file, ALBEDO_comp, geo, proj)
                        
            SE_root_file = os.path.join(folder_input_ETLook_Date, "se_root_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))
            if not os.path.exists(SE_root_file):
                
                input_format_SE_root = os.path.join(LS_WAPOR_folder_out, "{yyyy}{mm:02d}{dd:02d}", "se_root_{yyyy}{mm:02d}{dd:02d}.tif")
                SE_root_comp, geo, proj = Calc_Composite_Indicators_LS(input_format_SE_root, Date, Timestep_composite, method = "mean")
                DC.Save_as_tiff(SE_root_file, SE_root_comp, geo, proj)
    
            # Get example files
            dest_ex = gdal.Open(NDVI_file)
            geo_ex = dest_ex.GetGeoTransform()
            proj_ex = dest_ex.GetProjection()
            size_x_ex = dest_ex.RasterXSize
            size_y_ex = dest_ex.RasterYSize    

            ####################### Create lat and lon rasters ############################
            
            Lon_file = os.path.join(folder_input_ETLook_Static, "Lon.tif")     
            Lat_file = os.path.join(folder_input_ETLook_Static, "Lat.tif")            
            if not (os.path.exists(Lon_file) or os.path.exists(Lat_file)): 
                lon_deg = np.array([geo_ex[0] + np.arange(0,size_x_ex) * geo_ex[1]]*size_y_ex)
                lat_deg = np.array([geo_ex[3] + np.arange(0,size_y_ex) * geo_ex[5]]*size_x_ex).transpose()
                
                # save as tiff 
                DC.Save_as_tiff(Lon_file, lon_deg, geo_ex, proj_ex)
                DC.Save_as_tiff(Lat_file, lat_deg, geo_ex, proj_ex)
            
            else:
               dest_lon = gdal.Open(Lon_file)
               lon_deg = dest_lon.GetRasterBand(1).ReadAsArray()
               
            dlat, dlon = watertools.Functions.Area_Conversions.Area_converter.Calc_dlat_dlon(geo_ex, size_x_ex, size_y_ex)
            
            offset_GTM = int(round(lon_deg[int(lon_deg.shape[0]/2),int(lon_deg.shape[1]/2)] * 24 / 360))

            ########################## Create Time rasters ################################

            # Get DOY
            date_middle = Date + datetime.timedelta(days=int(Timestep_composite/2))             
            doy = int(date_middle.strftime("%j"))
 
            ####################### Create DEM rasters ############################
            
            # Create DEM files for ETLook            
            DEM_file = os.path.join(folder_input_ETLook_Static, "DEM.tif")                
            if not os.path.exists(DEM_file):
                folder_RAW_file_DEM = os.path.join(folders_input_RAW, "SRTM", "DEM")      
                filename_DEM = "DEM_SRTM_m_3s.tif"    
                if os.path.exists(os.path.join(folder_RAW_file_DEM, filename_DEM)):        
                    destDEM = RC.reproject_dataset_example(os.path.join(folder_RAW_file_DEM, filename_DEM), NDVI_file, method=4)
                    DEM = destDEM.GetRasterBand(1).ReadAsArray()
                    DC.Save_as_tiff(DEM_file, DEM, geo_ex, proj_ex)
                    
                else:
                    print("DEM is not available")        
        
            ##################### Calculate SLope and Aspect ##############################
            Slope_file = os.path.join(folder_input_ETLook_Static, "Slope.tif")                
            Aspect_file = os.path.join(folder_input_ETLook_Static, "Aspect.tif")   
            if not (os.path.exists(Slope_file) and os.path.exists(Aspect_file)):
                
                # open DEM
                destDEM = gdal.Open(DEM_file)
                DEM = destDEM.GetRasterBand(1).ReadAsArray()
                
                # constants
                pixel_spacing = (np.nanmean(dlon) +np.nanmean(dlat))/2
                deg2rad = np.pi / 180.0  # Factor to transform from degree to rad
                rad2deg = 180.0 / np.pi  # Factor to transform from rad to degree
            
                # Calculate slope
                x, y = np.gradient(DEM, pixel_spacing, pixel_spacing)
                hypotenuse_array = np.hypot(x,y)
                slope = np.arctan(hypotenuse_array) * rad2deg
            
                # calculate aspect
                aspect = np.arctan2(y/pixel_spacing, -x/pixel_spacing) * rad2deg
                aspect = 180 + aspect
        
                # Save as tiff files
                DC.Save_as_tiff(Slope_file, slope, geo_ex, proj_ex)
                DC.Save_as_tiff(Aspect_file, aspect, geo_ex, proj_ex)    
                
            ######################### Create Rainfall file ################################
            
            P_file = os.path.join(folder_input_ETLook_Date, "Precipitation_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))            
            if not os.path.exists(P_file): 
                
                folder_RAW_file_P = os.path.join(folders_input_RAW, "Precipitation", "CHIRPS", "Daily")                    

                input_format = os.path.join(folder_RAW_file_P, "P_CHIRPS.v2.0_mm-day-1_daily_{yyyy}.{mm:02d}.{dd:02d}.tif")
                P = pywapor.pre_et_look.Calc_Composite_METEO(input_format, Date, NDVI_file, "mean")
                P[np.isnan(P)] = 0
                DC.Save_as_tiff(P_file, P, geo_ex, proj_ex)                   


            ############################# Download METEO ##################################
            
            # Define the startdates for the METEO
            Date_end = Date + datetime.timedelta(days = int(Timestep_composite - 1)) 
            StartTime = datetime.datetime(Date.year, Date.month, Date.day, 0, 0)
            EndTime = datetime.datetime(Date_end.year, Date_end.month, Date_end.day, 23, 59)                                  
         
            # Download METEO data
            #if Date < datetime.datetime(2016,1,1): 
                # Test https://opendap.nccs.nasa.gov/dods/GEOS-5/MERRAero/hourly/tavg3hr_2d_asm_Nx.ascii?u10m[20:1:24][227:1:255][384:1:420]
            if Date < datetime.datetime(2000,1,1): 
                watertools.Collect.MERRA.daily(folders_input_RAW, ['u2m', 'v2m', 'q2m', 'tpw', 'ps', 'slp'],StartTime, EndTime, latlim, lonlim)
                watertools.Collect.MERRA.daily(folders_input_RAW, ['t2m'], StartTime, EndTime, latlim, lonlim, data_type = ["mean", "min", "max"])
                str_METEO = "MERRA"
                day_name = "daily"
                 
            #elif (Date >= datetime.datetime(2016,1,1) and Date < datetime.datetime(2017,12,1)):     
            elif (Date >= datetime.datetime(2000,1,1) and Date < datetime.datetime(2017,12,1)):     

                watertools.Collect.MERRA.daily_MERRA2(folders_input_RAW, ['u2m', 'v2m', 'q2m', 'tpw', 'ps', 'slp'],StartTime, EndTime, latlim, lonlim)
                watertools.Collect.MERRA.daily_MERRA2(folders_input_RAW, ['t2m'], StartTime, EndTime, latlim, lonlim, data_type = ["mean", "min", "max"])
                str_METEO = "MERRA"   
                day_name = "daily_MERRA2"
                
            else:
                watertools.Collect.GEOS.daily(folders_input_RAW, ['u2m', 'v2m', 'qv2m', 'tqv', 'ps', 'slp'], StartTime, EndTime, latlim, lonlim)
                watertools.Collect.GEOS.daily(folders_input_RAW, ['t2m'], StartTime, EndTime, latlim, lonlim, data_type = ["mean", "min", "max"])
                str_METEO = "GEOS"
                day_name = "daily"
                
            # Air pressure
            pair_24_0_file = os.path.join(folder_input_ETLook_Date, "Pair_24_0_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))        

            if not os.path.exists(pair_24_0_file):   
                
                folder_RAW_file_pair_24_0 = os.path.join(folders_input_RAW, str_METEO, "Sea_Level_Pressure", day_name) 
                
                input_format_pair_24_sea = os.path.join(folder_RAW_file_pair_24_0, "slp_%s_kpa_daily_{yyyy}.{mm:02d}.{dd:02d}.tif" %(str_METEO))   
  
                Pair_24_sea = pywapor.pre_et_look.Calc_Composite_METEO(input_format_pair_24_sea, Date, NDVI_file, "mean")
                Pair_24_sea[np.isnan(Pair_24_sea)] = 0     
             
                DC.Save_as_tiff(pair_24_0_file, Pair_24_sea, geo_ex, proj_ex)
                
            # Specific Humidity
            qv_24_file = os.path.join(folder_input_ETLook_Date, "qv_24_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))  
                     
            if not os.path.exists(qv_24_file):   
                folder_RAW_file_qv_24 = os.path.join(folders_input_RAW, str_METEO, "Specific_Humidity", day_name) 
                if str_METEO == "MERRA":
                    para = "q2m"
                else:
                    para = "qv2m"
 
                input_format_qv_24 = os.path.join(folder_RAW_file_qv_24, "%s_%s_kg-kg-1_daily_{yyyy}.{mm:02d}.{dd:02d}.tif" %(para, str_METEO))
   
                qv_24 = pywapor.pre_et_look.Calc_Composite_METEO(input_format_qv_24, Date, NDVI_file, "mean")
                qv_24[np.isnan(qv_24)] = 0                       
                
                DC.Save_as_tiff(qv_24_file, qv_24, geo_ex, proj_ex)

            # Air temperature
            Tair_24_file = os.path.join(folder_input_ETLook_Date, "tair_24_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))  
            Tair_max_24_file = os.path.join(folder_input_ETLook_Date, "tair_max_24_%d%02d%02d.tif" %(Date.year, Date.month, Date.day)) 
            Tair_min_24_file = os.path.join(folder_input_ETLook_Date, "tair_min_24_%d%02d%02d.tif" %(Date.year, Date.month, Date.day)) 
                     
            if not (os.path.exists(Tair_24_file) and os.path.exists(Tair_max_24_file) and os.path.exists(Tair_min_24_file)):
                folder_RAW_file_tair_24 = os.path.join(folders_input_RAW, str_METEO, "Air_Temperature", day_name) 
                
                input_format_tair_24 = os.path.join(folder_RAW_file_tair_24, "t2m_%s_K_daily_{yyyy}.{mm:02d}.{dd:02d}.tif" %(str_METEO))
                input_format_tair_min_24 = os.path.join(folder_RAW_file_tair_24,  "min", "t2mmin_%s_K_daily_{yyyy}.{mm:02d}.{dd:02d}.tif" %(str_METEO))
                input_format_tair_max_24 = os.path.join(folder_RAW_file_tair_24,  "max", "t2mmax_%s_K_daily_{yyyy}.{mm:02d}.{dd:02d}.tif" %(str_METEO))        

                tair_24 = pywapor.pre_et_look.Calc_Composite_METEO(input_format_tair_24, Date, NDVI_file, "max", DEM_file = DEM_file, lapse = -0.006)
                tair_24[np.isnan(tair_24)] = 0                       
                if np.nanmax(tair_24>270):
                    tair_24 = tair_24 -273.15
                    
                DC.Save_as_tiff(Tair_24_file, tair_24, geo_ex, proj_ex)

                tair_min_24 = pywapor.pre_et_look.Calc_Composite_METEO(input_format_tair_min_24, Date, NDVI_file, "max", DEM_file = DEM_file, lapse = 0.0)
                tair_min_24[np.isnan(tair_min_24)] = 0                       
                if np.nanmax(tair_min_24>270):
                    tair_min_24 = tair_min_24 -273.15
                                       
                DC.Save_as_tiff(Tair_min_24_file, tair_min_24, geo_ex, proj_ex)                   
                    
                tair_max_24 = pywapor.pre_et_look.Calc_Composite_METEO(input_format_tair_max_24, Date, NDVI_file, "max", DEM_file = DEM_file, lapse = 0.0)
                tair_max_24[np.isnan(tair_max_24)] = 0                       
                if np.nanmax(tair_max_24>270):
                    tair_max_24 = tair_max_24 -273.15                       
                DC.Save_as_tiff(Tair_max_24_file, tair_max_24, geo_ex, proj_ex)                   
        
            # Wind Speed
            wind_24_file = os.path.join(folder_input_ETLook_Date, "wind_24_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))  
                     
            if not os.path.exists(wind_24_file):
                     
                folder_RAW_file_u2_24 = os.path.join(folders_input_RAW, str_METEO, "Eastward_Wind", day_name)      
                folder_RAW_file_v2_24 = os.path.join(folders_input_RAW, str_METEO, "Northward_Wind", day_name) 

                input_format_u2_24 = os.path.join(folder_RAW_file_u2_24, "u2m_%s_m-s-1_daily_{yyyy}.{mm:02d}.{dd:02d}.tif" %(str_METEO))
                input_format_v2_24 = os.path.join(folder_RAW_file_v2_24, "v2m_%s_m-s-1_daily_{yyyy}.{mm:02d}.{dd:02d}.tif" %(str_METEO))

                u2_24= pywapor.pre_et_look.Calc_Composite_METEO(input_format_u2_24, Date, NDVI_file, "mean")
                u2_24[np.isnan(u2_24)] = 0                       
 
                v2_24= pywapor.pre_et_look.Calc_Composite_METEO(input_format_v2_24, Date, NDVI_file, "mean")
                v2_24[np.isnan(v2_24)] = 0    
                
                wind_24 = np.sqrt(u2_24**2 + v2_24 **2) 
                 
                DC.Save_as_tiff(wind_24_file, wind_24, geo_ex, proj_ex)                      

                    
            ##################### Calculate Landmask ##############################
                      
            LM_file = os.path.join(folder_input_ETLook_Static, "LandMask_%s.tif" %Date.year)   
            Bulk_file = os.path.join(folder_input_ETLook_Static, "Bulk_Stomatal_resistance_%s.tif" %Date.year)   
            MaxObs_file = os.path.join(folder_input_ETLook_Static, "Maximum_Obstacle_Height_%s.tif" %Date.year)  
            LUEmax_file = os.path.join(folder_input_ETLook_Static, "LUEmax_%s.tif" %Date.year)  
            if not (os.path.exists(LM_file) and os.path.exists(Bulk_file) and os.path.exists(MaxObs_file) and os.path.exists(LUEmax_file)):
                
                if LandCover == "GlobCover":
                    folder_RAW_file_LC = os.path.join(folders_input_RAW, "GlobCover", "Landuse")      
                    filename_LC = "LC_GLOBCOVER_V2.3.tif"    
                if LandCover == "WAPOR":
                    folder_RAW_file_LC = os.path.join(folders_input_RAW, "L1_LCC_A")      
                    filename_LC = "L1_LCC_A_WAPOR_YEAR_%s.01.01.tif" %(Date.year)                      
                    
                if os.path.exists(os.path.join(folder_RAW_file_LC, filename_LC)):        
                    destLC = RC.reproject_dataset_example(os.path.join(folder_RAW_file_LC, filename_LC), NDVI_file, method=1)
                    LC = destLC.GetRasterBand(1).ReadAsArray()
                    LC[np.isnan(LC)] = -9999
                    
                    # import list with numbers to convert globcover into other maps
                    import pywapor.Functions.LandCover_Converter as LCC
                    
                    if LandCover == "GlobCover":
                        # Get conversion between globcover and landmask
                        LU_LM_Classes = LCC.Globcover_LM()
                        LU_Bulk_Classes = LCC.Globcover_Bulk()
                        LU_MaxObs_Classes = LCC.Globcover_MaxObs()
                        LU_LUEmax_Classes = LCC.Globcover_LUEmax()
                        
                    if LandCover == "WAPOR":
                        # Get conversion between globcover and landmask
                        LU_LM_Classes = LCC.WAPOR_LM()
                        LU_Bulk_Classes = LCC.WAPOR_Bulk()
                        LU_MaxObs_Classes = LCC.WAPOR_MaxObs()    
                        LU_LUEmax_Classes = LCC.WAPOR_LUEmax()
                        
                    # Create Array for LandMask
                    LM = np.ones([size_y_ex, size_x_ex]) * np.nan            
                    Bulk = np.ones([size_y_ex, size_x_ex]) * np.nan 
                    MaxObs = np.ones([size_y_ex, size_x_ex]) * np.nan 
                    LUEmax = np.ones([size_y_ex, size_x_ex]) * np.nan 
                     
                    # Create LandMask 
                    for LU_LM_Class in LU_LM_Classes.keys():
                        Value_LM = LU_LM_Classes[LU_LM_Class]
                        Value_Bulk = LU_Bulk_Classes[LU_LM_Class]
                        Value_MaxObs = LU_MaxObs_Classes[LU_LM_Class]                
                        Value_LUEmax = LU_LUEmax_Classes[LU_LM_Class]   
                        LM[LC == LU_LM_Class] = Value_LM
                        Bulk[LC == LU_LM_Class] = Value_Bulk
                        MaxObs[LC  == LU_LM_Class] = Value_MaxObs
                        LUEmax[LC  == LU_LM_Class] = Value_LUEmax
                        
                    # Save as tiff files               
                    DC.Save_as_tiff(LM_file, LM, geo_ex, proj_ex)
                    DC.Save_as_tiff(Bulk_file, Bulk, geo_ex, proj_ex)
                    DC.Save_as_tiff(MaxObs_file, MaxObs, geo_ex, proj_ex)
                    DC.Save_as_tiff(LUEmax_file, LUEmax, geo_ex, proj_ex)
                    
                else:
                    print("LandCover is not available")       
            
            ########################### Download amplitude ################################
                
            # yearly amplitude temperature air
            Tair_amp_file = os.path.join(folder_input_ETLook_Static, "Tair_amp_%d.tif" %(Date.year))     
            output_folder_Tamp = os.path.join(folders_input_RAW, "GLDAS")       
            
            if not os.path.exists(Tair_amp_file):
                T_amplitude_global_temp_filename = os.path.join(output_folder_Tamp, "Temp_Amplitudes_global.tif")
                if os.path.exists(T_amplitude_global_temp_filename):        
                    desttairamp = RC.reproject_dataset_example(T_amplitude_global_temp_filename, NDVI_file, method=2)
                    tair_amp = desttairamp.GetRasterBand(1).ReadAsArray()
                    DC.Save_as_tiff(Tair_amp_file, tair_amp, geo_ex, proj_ex)
                    
                else:
                    print("Yearly Tair amplitude is not available")    
        
            ######################## Download Transmissivity ##############################
            
            # Download MSGCPP data
            #if Date < datetime.datetime(2016,1,1): 
            if Date < datetime.datetime(2002,1,1): 
                watertools.Collect.MERRA.daily(folders_input_RAW, ['swgnet'],StartTime, EndTime, latlim, lonlim)
                str_TRANS = "MERRA"
                day_name = "daily"

            elif (Date >= datetime.datetime(2002,1,1) and Date < datetime.datetime(2017,1,1)):        
            #elif (Date >= datetime.datetime(2016,1,1) and Date < datetime.datetime(2017,1,1)):
                watertools.Collect.MERRA.daily_MERRA2(folders_input_RAW, ['swgnet'],StartTime, EndTime, latlim, lonlim)
                str_TRANS = "MERRA"    
                day_name = "daily_MERRA2"
                
            else:
                if Short_Downwards_Radiation ==  "MSGCCP":
                    watertools.Collect.MSGCPP.SDS(folders_input_RAW, StartTime, EndTime, latlim, lonlim)
                    str_TRANS = "MSGCPP"
                    day_name = "daily"
                if Short_Downwards_Radiation == "MERRA":
                    watertools.Collect.MERRA.daily_MERRA2(folders_input_RAW, ['swgnet'],StartTime, EndTime, latlim, lonlim)
                    str_TRANS = "MERRA"    
                    day_name = "daily_MERRA2"            
                    
            # yearly amplitude temperature air
            Trans_file = os.path.join(folder_input_ETLook_Date, "Trans_24_%d%02d%02d.tif" %(Date.year, Date.month, Date.day))   
                     
            if not os.path.exists(Trans_file):
                
                # Calculate the extraterrestrial daily radiation
                destLat = gdal.Open(Lat_file)
                lat = destLat.GetRasterBand(1).ReadAsArray()
                
                Gsc = 1367        # Solar constant (W / m2)
                deg2rad = np.pi / 180.0
                # Computation of Hour Angle (HRA = w)
                B = 360./365 * (doy-81)           # (degrees)
                # Computation of cos(theta), where theta is the solar incidence angle
                # relative to the normal to the land surface
                delta=np.arcsin(np.sin(23.45*deg2rad)*np.sin(np.deg2rad(B))) # Declination angle (radians)
                
                phi = lat * deg2rad                                     # latitude of the pixel (radians)
                
                dr = 1 + 0.033 * np.cos(doy*2*np.pi/365)
                
                # Daily 24 hr radiation - For flat terrain only !
                ws_angle = np.arccos(-np.tan(phi)*np.tan(delta))   # Sunset hour angle ws   
                
                # Extraterrestrial daily radiation, Ra (W/m2):
                Ra24_flat = (Gsc/np.pi * dr * (ws_angle * np.sin(phi) * np.sin(delta) +
                                np.cos(phi) * np.cos(delta) * np.sin(ws_angle)))
                    
                Dates_End_Rads = Date + datetime.timedelta(days = int(Timestep_composite - 1)) 
                Dates_rads = pd.date_range(Date, Dates_End_Rads)
                i = 0
                for Date_rad in Dates_rads:
                    if str_TRANS == "MERRA":
                        folder_RAW_file_trans = os.path.join(folders_input_RAW, str_TRANS, "Surface_Net_Downward_Shortwave_Flux", day_name) 
                        filename_trans = "swgnet_MERRA_W-m-2_daily_%d.%02d.%02d.tif"  %(Date.year, Date.month, Date.day)            
                    if str_TRANS == "MSGCPP":
                        folder_RAW_file_trans = os.path.join(folders_input_RAW, str_TRANS, "SDS", "15min") 
                        filename_trans = "SDS_MSGCPP_W-m-2_15min_%d.%02d.%02d_H{hour}.M{minutes}.tif"  %(Date.year, Date.month, Date.day)
                        
                        os.chdir(folder_RAW_file_trans)
                        files = glob.glob(filename_trans.format(hour = "*", minutes = "*"))
                        i = 0
                        
                        # Open all the 15 minutes files
                        for file in files:
                            file_in = os.path.join(folder_RAW_file_trans, file)
                            destswgone = gdal.Open(file_in)
                            try:
                                swgnet_one = destswgone.GetRasterBand(1).ReadAsArray()
                                swgnet_one[swgnet_one<0] = 0                 
                                if not "geo_trans" in locals():
                                    swgnet = np.ones([destswgone.RasterYSize, destswgone.RasterXSize, len(files)]) * np.nan
                                    geo_trans = destswgone.GetGeoTransform()
                                    proj_trans = destswgone.GetProjection()
                                swgnet[:,:,i] = swgnet_one
                            except:
                                pass
                            i+=1 
                        
                        # Calculate the daily mean     
                        swgnet_mean = np.nanmean(swgnet, 2)
                        dest_swgnet_mean = DC.Save_as_MEM(swgnet_mean,geo_trans, proj_trans)
                        destswgnet = RC.reproject_dataset_example(dest_swgnet_mean, NDVI_file, method=2)
                        del geo_trans
                        
                    else:
                        destswgnet = RC.reproject_dataset_example(os.path.join(folder_RAW_file_trans, filename_trans), NDVI_file, method=2)
                         
                    swgnet_one = destswgnet.GetRasterBand(1).ReadAsArray()
                    if i == 0:
                        swgnet = np.ones([8, destswgnet.RasterYSize, destswgnet.RasterXSize])
                    swgnet[i, :, :] = swgnet_one
                    i+=1
                    
                trans = np.nanmean(swgnet,axis = 0) / Ra24_flat
                DC.Save_as_tiff(Trans_file, trans, geo_ex, proj_ex)

                
        except Exception as e:
            print("No ETLook input dataset for %s" %Date)
            print(e) 
            
    return()


def Calc_Composite_Indicators_LS(input_format, Date, Timestep_composite, method = "mean"):
    
    # Find dates LS
    Dates_LS = np.array([datetime.datetime.strptime(k, "%Y%m%d").toordinal() for k in os.listdir(os.path.dirname(os.path.dirname(input_format))) if len(k)==8])
    Dates_LS_selected = Dates_LS[np.logical_and(Dates_LS>=Date.toordinal(), Dates_LS<=(Date + datetime.timedelta(days=int(Timestep_composite-1))).toordinal())]
    
    i = 0
    for Date_LS_selected in Dates_LS_selected:
        date_comp = datetime.datetime.fromordinal(Date_LS_selected)
        print("Find %s" %input_format.format(yyyy=date_comp.year, mm=date_comp.month, dd=date_comp.day))
        filename = glob.glob(input_format.format(yyyy=date_comp.year, mm=date_comp.month, dd=date_comp.day))[0]  
        
        dest_one = gdal.Open(filename)
        Array = dest_one.GetRasterBand(1).ReadAsArray()
        Array[Array==-9999] = np.nan
        proj_filename = filename
    
        if Date_LS_selected == Dates_LS_selected[0]:
            dest_one = gdal.Open(proj_filename)
            geo = dest_one.GetGeoTransform()
            proj = dest_one.GetProjection()
            Array_end = np.ones([int(Timestep_composite), dest_one.RasterYSize, dest_one.RasterXSize]) * np.nan
        
        Array_end[i, :, :] = Array
        i+=1

    if method == "mean":
        Array_end = np.nanmean(Array_end, axis = 0)
    if method == "max":
        Array_end = np.nanmax(Array_end, axis = 0)        
    if method == "min":
        Array_end = np.nanmin(Array_end, axis = 0)
    if method == "sum":
        Array_end = np.nansum(Array_end, axis = 0)
        
    return(Array_end, geo, proj)